# Copyright 2024 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file contains a Megatron style Hybrid Engine that shares the weights of the actor with the inference engine.
"""

import inspect

import torch
import torch.distributed
import torch.distributed as dist
from megatron.core import DistributedDataParallel as LocalDDP
from megatron.core import parallel_state as mpu
from megatron.core.transformer.module import Float16Module
from torch import nn
from torch.nn.parallel.distributed import DistributedDataParallel as torchDDP

from siirl import DataProto
from siirl.models.mcore.weight_converter import McoreToHFWeightConverterBase
from siirl.workers.databuffer import all_gather_data_proto
from vllm import LLM
from vllm.distributed import parallel_state as vllm_ps
from siirl.utils.debug import GPUMemoryLogger
from siirl.utils.extras.device import get_torch_device
from siirl.utils.megatron.megatron_utils import (
    get_model,
    per_tensor_generator,
    unwrap_model,
)
from siirl.utils.megatron.memory_buffer import (
    build_memory_buffer,
    build_memory_reference_from_module,
    get_weight_buffer_meta_from_module,
)
from siirl.utils.model_utils.torch_functional import check_device_is_available
from siirl.utils.model_utils.vllm_utils import patch_vllm_moe_model_weight_loader

from siirl.workers.sharding_manager.base import BaseShardingManager

from loguru import logger


class AllGatherPPModel:
    def __init__(self, model_provider, use_distributed_optimizer=True) -> None:
        print(
            "[WARNING] This class is deprecated and will no longer be supported. \
Consider using the `MegatronPPOActor` class directly as a replacement."
        )
        self._pp_group = mpu.get_pipeline_model_parallel_group()
        self._pp_rank = mpu.get_pipeline_model_parallel_rank()
        self._pp_size = mpu.get_pipeline_model_parallel_world_size()
        self._vpp_size = mpu.get_virtual_pipeline_model_parallel_world_size()
        self._model_chunk_size = self._vpp_size or 1

        # each one holds a list of model_chunks in this pp stage
        self._pp_models = [None] * self.pp_size

        rank_list = list(range(self.pp_size))
        # make current rank the last one to initialize
        rank_list[self.pp_rank], rank_list[-1] = rank_list[-1], rank_list[self.pp_rank]
        self._this_rank_models = None

        # store the parameter of each pp stage
        self.memory_buffers = [None] * self.pp_size
        for cur_pp_rank in rank_list:
            print(
                "create pp model",
                f"torch allocated {torch.cuda.memory_allocated() / 1e9:.4f} GB, reserved {torch.cuda.memory_reserved() / 1e9:.4f} GB",
            )
            # since the last initialized rank is the current pp rank, after init, the pp rank is still correct
            mpu.set_pipeline_model_parallel_rank(cur_pp_rank)
            if cur_pp_rank != self.pp_rank:
                models = get_model(model_provider, wrap_with_ddp=False, use_distributed_optimizer=False)
                models = nn.ModuleList(models)
                assert len(models) == self._model_chunk_size, f"{len(models)} != {self._model_chunk_size}"
                self.pp_models[cur_pp_rank] = models
            else:
                # for regular model, we wrapped it with DDP
                models = get_model(model_provider, wrap_with_ddp=True, use_distributed_optimizer=use_distributed_optimizer)
                assert len(models) == self._model_chunk_size, f"{len(models)} != {self._model_chunk_size}"
                self._this_rank_models = nn.ModuleList(models)
                self.pp_models[cur_pp_rank] = nn.ModuleList(unwrap_model(models, (torchDDP, LocalDDP)))

            self._build_param_buffer(cur_pp_rank)
            self._build_param_references(cur_pp_rank, maintain_weight=cur_pp_rank == self.pp_rank)

            # TODO: after binding to the memory buffer, we can load the checkpoint here
            if cur_pp_rank != self.pp_rank:
                for model in self.pp_models[cur_pp_rank]:
                    model.eval()
                self._offload_params_to_cpu(cur_pp_rank)

    def _build_param_buffer(self, pp_rank):
        """Build the parameter buffer in each pp rank"""
        if pp_rank == self._pp_rank:
            from siirl.utils.megatron.memory_buffer import MemoryBuffer

            # The code here is very hard-coded, based on the following assumptions:
            # 1. `len(_this_rank_models) == 1`
            # 2. `_this_rank_models[0]` is a instance of `DistributedDataParallel` and `use_distributed_optimizer=True`
            # 3. Only bfloat16 data type is used in parameters
            source = self._this_rank_models[0].buffers[0].param_data
            self.memory_buffers[pp_rank] = {torch.bfloat16: MemoryBuffer(source.numel(), source.numel(), torch.bfloat16, source)}
        else:
            model = self.pp_models[pp_rank]
            weight_buffer_meta = get_weight_buffer_meta_from_module(model)
            self.memory_buffers[pp_rank] = build_memory_buffer(weight_buffer_meta)

    def _build_param_references(self, pp_rank, maintain_weight=False):
        if pp_rank == self._pp_rank:
            return
        model = self.pp_models[pp_rank]
        build_memory_reference_from_module(model, self.memory_buffers[pp_rank], maintain_weight=maintain_weight)

    def _load_params_to_cuda(self, pp_rank, to_empty=False):
        assert pp_rank != self.pp_rank, f"unexpected to load current pp rank [{pp_rank}] back to cuda"
        for buffer in self.memory_buffers[pp_rank].values():
            if not to_empty:
                buffer.data = buffer.data.to(torch.cuda.current_device(), non_blocking=True)
            else:
                buffer.data = torch.empty_like(buffer.data, device="cuda")
        # rebuild reference after loading to CUDA
        self._build_param_references(pp_rank)

    def _offload_params_to_cpu(self, pp_rank, to_empty=False):
        assert pp_rank != self.pp_rank, f"unexpected to offload current pp rank [{pp_rank}] to cpu"
        for buffer in self.memory_buffers[pp_rank].values():
            if not to_empty:
                # offload the whole memory buffer to CPU
                buffer.data = buffer.data.to("cpu", non_blocking=True)
            else:
                buffer.data = torch.empty_like(buffer.data, device="cpu")
        self._build_param_references(pp_rank)

    def load_params_to_cuda(self, to_empty=False):
        """load all model params to cuda"""
        for cur_pp_rank in range(self.pp_size):
            if cur_pp_rank != self.pp_rank:
                self._load_params_to_cuda(cur_pp_rank, to_empty=to_empty)

    def allgather_params(self):
        """allgather params of all pp ranks. Return a list of handles"""
        for cur_pp_rank in range(self.pp_size):
            global_src = dist.get_global_rank(group=self.pp_group, group_rank=cur_pp_rank)

            # NOTE(sgm): the async op may cause memory leakage of the memory_buffer/pp_models

            for _, param in sorted(self.pp_models[cur_pp_rank].named_parameters()):
                dist.broadcast(tensor=param.data, src=global_src, group=self.pp_group, async_op=False)

    def forward(self, *inputs, **kwargs):
        try:
            prev_output = None
            for cur_chunk_rank in range(self._model_chunk_size):
                if self._vpp_size:
                    mpu.set_virtual_pipeline_model_parallel_rank(cur_chunk_rank)

                for cur_pp_rank in range(self.pp_size):
                    mpu.set_pipeline_model_parallel_rank(cur_pp_rank)
                    self.pp_models[cur_pp_rank][cur_chunk_rank].set_input_tensor(prev_output)
                    ret = self.pp_models[cur_pp_rank][cur_chunk_rank](*inputs, **kwargs)
                    self.pp_models[cur_pp_rank][cur_chunk_rank].set_input_tensor(None)
                    prev_output = ret
        finally:
            if self._vpp_size:
                mpu.set_virtual_pipeline_model_parallel_rank(0)
            mpu.set_pipeline_model_parallel_rank(self.pp_rank)
        return ret

    def __call__(self, *inputs, **kwargs):
        return self.forward(*inputs, **kwargs)

    def eval(self):
        for model in self.pp_models[self.pp_rank]:
            model.eval()

    def train(self):
        for model in self.pp_models[self.pp_rank]:
            model.train()

    def offload_params_to_cpu(self, to_empty=False):
        """offload params of models that are not of current pp rank to cpu"""
        for cur_pp_rank in range(self.pp_size):
            if cur_pp_rank != self.pp_rank:
                self._offload_params_to_cpu(cur_pp_rank, to_empty=to_empty)

    def get_all_params(self):
        """Get all the parameters of the models in all pp ranks

        Returns:
            params: List[List[Dict[str, Tensor]]]: a list of parameters in all pp, where each is a list of dict
                tensors of each model chunk

        """
        params = []
        for pp_rank in range(self.pp_size):
            params.append([])
            for model_chunk_idx in range(len(self.pp_models[pp_rank])):
                params[pp_rank].append({})
                pp_model = self.pp_models[pp_rank][model_chunk_idx]
                pp_model = unwrap_model(pp_model, ((torchDDP, LocalDDP, Float16Module)))  # not use Float16Module
                for name, param in pp_model.named_parameters():
                    # NOTE(gh) workaround: should not get lora params for inference
                    if "lora" in name:
                        continue
                    params[pp_rank][model_chunk_idx][name] = param

        return params

    def update_this_rank_models(self, new_models):
        self._this_rank_models = new_models
        self._pp_models[self.pp_rank] = unwrap_model(new_models, (torchDDP, LocalDDP))

    @property
    def this_rank_models(self):
        return self._this_rank_models

    @property
    def pp_size(self):
        return self._pp_size

    @property
    def pp_rank(self):
        return self._pp_rank

    @property
    def pp_group(self):
        return self._pp_group

    @property
    def pp_models(self):
        return self._pp_models


"""
Megatron Hybrid Engine:
- During training, only the current pp stage holds the parameters
- Before inference, broadcast the parameters of the current pp rank 
   to all other pp ranks (all pp ranks holds all the parameters)
- Bind the parameters to the inference engine
- Do inference in tp. pp is treated as additional dp
- After inference, all the parameters that doesn't belong to this pp rank is freed.
"""


# Micro Data parallel group. Micro data parallel group is additional dp group that origins from splitting training tp
# into infer_tp and micro_tp. By default, we use order micro_dp - tp
# NOTICE: in new version of vLLM, We need to all-gather all tp rank's model weights
# For code reuse, we directly assign Megatron's TENSOR_MODEL_PARALLEL_GROUP to this
_MICRO_DATA_PARALLEL_GROUP = None


class MegatronVLLMShardingManager(BaseShardingManager):
    @check_device_is_available()
    def __init__(
        self,
        actor_module: nn.ModuleList,
        inference_engine: LLM,
        model_config,
        transformer_config,
        layer_name_mapping,
        weight_converter: McoreToHFWeightConverterBase,
        module: AllGatherPPModel = None,
    ):
        from megatron.core import parallel_state as mpu

        self.actor_module = actor_module
        self.inference_engine = inference_engine
        self.model_config = model_config
        self.transformer_config = transformer_config
        self.layer_name_mapping = layer_name_mapping
        self.weight_converter = weight_converter
        self.module = module
        # initialize groups for vllm inference
        self.rank = torch.distributed.get_rank()
        self.world_size = torch.distributed.get_world_size()
        self.infer_tp_size = vllm_ps.get_tensor_model_parallel_world_size()
        self.infer_tp_rank = vllm_ps.get_tensor_model_parallel_rank()
        self.infer_tp_group = vllm_ps.get_tensor_model_parallel_group()
        self.infer_tp_group = self.infer_tp_group.device_group
        self.train_tp_size = mpu.get_tensor_model_parallel_world_size()
        self.train_tp_rank = mpu.get_tensor_model_parallel_rank()
        self.train_tp_group = mpu.get_tensor_model_parallel_group()
        self.train_ep_size = mpu.get_expert_model_parallel_world_size()
        self.train_ep_rank = mpu.get_expert_model_parallel_rank()
        self.train_ep_group = mpu.get_expert_model_parallel_group()
        self.train_etp_size = mpu.get_expert_tensor_parallel_world_size()
        self.train_etp_rank = mpu.get_expert_tensor_parallel_rank()
        self.train_etp_group = mpu.get_expert_tensor_parallel_group()
        self.need_tp_reshard = self.train_tp_size != self.infer_tp_size
        self.train_tp_larger = self.train_tp_size > self.infer_tp_size

    @GPUMemoryLogger(role="megatron vllm sharding_manager", logger=logger)
    def __enter__(self):
        # vllm > 0.7.2
        if "tags" in inspect.signature(self.inference_engine.wake_up).parameters:
            self.inference_engine.wake_up(tags=["weights"])
        else:
            self.inference_engine.wake_up()
        per_tensor_param = per_tensor_generator(
            self.actor_module,
            self.model_config,
            self.weight_converter,
            self.transformer_config,
            self.layer_name_mapping,
        )
        model = self.inference_engine.llm_engine.model_executor.driver_worker.worker.model_runner.model
        patch_vllm_moe_model_weight_loader(model)
        loaded_params = model.load_weights(per_tensor_param)
        info = f"vLLM load weights, loaded_params: {len(loaded_params)}"
        logger.info(info)

        # (vermouth1992) We move wake up kv cache after we release model weights. Need refactor to make API cleaner
        # if "tags" in inspect.signature(self.inference_engine.wake_up).parameters:
        #     self.inference_engine.wake_up(tags=["kv_cache"])

    @GPUMemoryLogger(role="megatron vllm sharding_manager", logger=logger)
    def __exit__(self, exc_type, exc_value, traceback):
        self.inference_engine.sleep(level=1)
        for model in self.actor_module:
            model.train()

        get_torch_device().empty_cache()

    @GPUMemoryLogger(role="megatron vllm sharding_manager", logger=logger)
    def preprocess_data(self, data: DataProto) -> DataProto:
        # DP_COMPUTE_PROTO: all training ranks are dp, the same as fsdp
        if self.infer_tp_size == 1:
            return data
        all_gather_data_proto(data, self.infer_tp_group)
        return data

    @GPUMemoryLogger(role="megatron vllm sharding_manager", logger=logger)
    def postprocess_data(self, data: DataProto) -> DataProto:
        # DP_COMPUTE_PROTO: all training ranks are dp, the same as fsdp
        if self.infer_tp_size == 1:
            return data
        return data.chunk(chunks=self.infer_tp_size)[self.infer_tp_rank]



from typing import TYPE_CHECKING, Any, Dict, Optional, TypedDict

from transformers import (
    AutoConfig,
    AutoProcessor,
    AutoTokenizer,
)

from siirl.utils.extras.misc import skip_check_imports, try_download_model_from_other_hub
from siirl.models.patcher import patch_processor, patch_tokenizer

from loguru import logger


if TYPE_CHECKING:
    from transformers import PreTrainedTokenizer, ProcessorMixin

    from siirl.utils.params import ModelArguments


class TokenizerModule(TypedDict):
    tokenizer: "PreTrainedTokenizer"
    processor: Optional["ProcessorMixin"]


def _get_init_kwargs(model_args: "ModelArguments") -> Dict[str, Any]:
    r"""
    Gets arguments to load config/tokenizer/model.
    Note: including inplace operation of model_args.
    """
    skip_check_imports()
    model_args.path = try_download_model_from_other_hub(model_args)
    return {
        "trust_remote_code": model_args.trust_remote_code,
        "cache_dir": model_args.cache_dir,
        "revision": model_args.model_revision,
        "token": model_args.hf_hub_token,
    }


def set_pad_token_id(tokenizer):
    """Set pad_token_id to eos_token_id if it is None.

    Args:
        tokenizer (transformers.PreTrainedTokenizer): The tokenizer to be set.

    """
    if tokenizer.pad_token_id is None:
        tokenizer.pad_token_id = tokenizer.eos_token_id
        logger.warning(f"tokenizer.pad_token_id is None. Now set to {tokenizer.eos_token_id}")
    if tokenizer.pad_token is None:
        tokenizer.pad_token = tokenizer.eos_token
        logger.warning(f"tokenizer.pad_token is None. Now set to {tokenizer.eos_token}")


def load_tokenizer(
    path: str = "",
    model_args: "ModelArguments" = None,
    correct_pad_token: bool = True,
    correct_gemma2=True,
) -> "TokenizerModule":
    r"""
    Loads pretrained tokenizer and optionally loads processor.
    Note: including inplace operation of model_args.
    """
    init_kwargs = {}
    if model_args is not None:
        path = model_args.path
        init_kwargs = _get_init_kwargs(model_args)
    config = AutoConfig.from_pretrained(path, **init_kwargs)
    if correct_gemma2 and isinstance(path, str) and "gemma-2-2b-it" in path:
        # the EOS token in gemma2 is ambiguious, which may worsen RL performance.
        # https://huggingface.co/google/gemma-2-2b-it/commit/17a01657f5c87135bcdd0ec7abb4b2dece04408a
        logger.warning("Found gemma-2-2b-it tokenizer. Set eos_token and eos_token_id to <end_of_turn> and 107.", stacklevel=1)
        init_kwargs["eos_token"] = "<end_of_turn>"
        init_kwargs["eos_token_id"] = 107
    try:
        if "InternVL" in config.architectures[0] and "internlm2" in config.llm_config.model_type:
            from siirl.models.transformers.internvl_chat.tokenization_internlm2_fast import InternLM2TokenizerFast

            tokenizer = InternLM2TokenizerFast.from_pretrained(
                path,
                use_fast=True,
                split_special_tokens=model_args.split_special_tokens if model_args else False,
                padding_side="right",
                **init_kwargs,
            )
        else:
            tokenizer = AutoTokenizer.from_pretrained(
                path,
                use_fast=model_args.use_fast_tokenizer if model_args else True,
                split_special_tokens=model_args.split_special_tokens if model_args else False,
                padding_side="right",
                **init_kwargs,
            )
    except ValueError:  # try the fast one
        tokenizer = AutoTokenizer.from_pretrained(
            path,
            use_fast=True,
            padding_side="right",
            **init_kwargs,
        )
    except Exception as e:
        raise OSError("Failed to load tokenizer.") from e

    if model_args:
        patch_tokenizer(tokenizer, model_args, config)
    try:
        processor = AutoProcessor.from_pretrained(path, **init_kwargs)
        if model_args:
            patch_processor(processor, config, tokenizer, model_args)
    except Exception as e:
        logger.debug(f"Processor was not found: {e}.")
        processor = None

    # Avoid load tokenizer, see:
    # https://github.com/huggingface/transformers/blob/v4.40.0/src/transformers/models/auto/processing_auto.py#L324
    if processor is not None and "Processor" not in processor.__class__.__name__:
        processor = None

    if processor is None and "InternVL" in config.architectures[0]:
        import torch
        from siirl.models.transformers.internvl import build_transform, dynamic_preprocess

        class InternVLProcessor:
            def __init__(self, proc, config):
                self.proc = proc
                self.image_size = config.force_image_size or config.vision_config.image_size
                self.patch_size = config.vision_config.patch_size
                self.dynamic_image_size = False  # config.dynamic_image_size
                self.min_dynamic_patch = config.min_dynamic_patch
                self.max_dynamic_patch = config.max_dynamic_patch
                self.use_thumbnail = config.use_thumbnail
                self.num_image_token = int((self.image_size // self.patch_size) ** 2 * (config.downsample_ratio**2))

            def process(self, images):
                transform = self.proc(True, input_size=self.image_size)
                pixel_values, image_flags = [], []
                for image in images:
                    pixel_values.append(transform(image))
                    image_flags.append(torch.tensor([1] * 1, dtype=torch.long))
                pixel_values = torch.stack(pixel_values)
                image_flags = torch.stack(image_flags)
                return {"pixel_values": pixel_values, "image_flags": image_flags}

        processor = InternVLProcessor(build_transform, config)
        logger.info("Build processor for model type ", config.model_type)

    if correct_pad_token:
        set_pad_token_id(tokenizer)

    return {"tokenizer": tokenizer, "processor": processor}



# Copyright 2025, Shanghai Innovation Institute. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time

import hydra
import ray
from loguru import logger
from omegaconf import DictConfig

from siirl.workers.dag import DAGConfigLoader
from siirl.workers.databuffer import init_data_buffer
from siirl.scheduler.launch import RayTrainer
from siirl.scheduler.graph_updater import display_node_config
from siirl.utils.params import log_dict_formatted
from siirl.scheduler.graph_updater import update_task_graph_node_configs
from siirl.scheduler.process_group_manager import ProcessGroupManager, log_process_group_manager_details
from siirl.scheduler.task_scheduler import TaskScheduler, log_schedule_assignments
from siirl.utils.logger.logging_utils import set_basic_config
from siirl.utils.params import SiiRLArguments, parse_config
from siirl.workers.environment import EnvironmentConfigLoader
from siirl.scheduler.enums import AdvantageEstimator

# --- Constants ---
RAY_RUNTIME_ENV_VARS = {
    "TOKENIZERS_PARALLELISM": "true",
    "NCCL_DEBUG": "WARN",
    "VLLM_LOGGING_LEVEL": "WARN",
}

# The main runner is an orchestrator, not a heavy workload.
# Assigning it a full CPU is often wasteful. A fractional CPU is more efficient.
MAIN_RUNNER_CPU_RESERVATION = 5


def determine_workflow_config(self, siirl_args: SiiRLArguments) -> str:
    current_dir = os.path.dirname(os.path.abspath(__file__))

    if siirl_args.algorithm.adv_estimator == AdvantageEstimator.GAE:
        return os.path.join(current_dir, "config/workflow_ppo.yaml")
    elif siirl_args.algorithm.adv_estimator in [
        AdvantageEstimator.GRPO,
        AdvantageEstimator.GRPO_PASSK,
        AdvantageEstimator.REINFORCE_PLUS_PLUS,
        AdvantageEstimator.REMAX,
        AdvantageEstimator.RLOO,
        AdvantageEstimator.OPO,
        AdvantageEstimator.REINFORCE_PLUS_PLUS_BASELINE,
        AdvantageEstimator.CPGD,
    ]:
        return os.path.join(current_dir, "config/workflow_grpo.yaml")
    else:
        raise NotImplementedError


def get_databuffer_shard_number(siirl_args: SiiRLArguments) -> int:
    assert siirl_args.data.train_batch_size % siirl_args.trainer.nnodes == 0, f"Config Error: train_batch_size ({siirl_args.data.train_batch_size}) must be divisible by nnodes ({siirl_args.trainer.nnodes}). Please adjust your configuration."

    batch_size_per_node = siirl_args.data.train_batch_size // siirl_args.trainer.nnodes
    intermediate_value = batch_size_per_node * siirl_args.actor_rollout_ref.rollout.n
    SHARDING_FACTOR = 8
    assert intermediate_value % SHARDING_FACTOR == 0, f"Config Error: The result of '(train_batch_size / nnodes) * rollout_n' ({intermediate_value}) must be divisible by {SHARDING_FACTOR}. Please adjust your configuration."
    databuffer_number = ((siirl_args.data.train_batch_size // siirl_args.trainer.nnodes) * siirl_args.actor_rollout_ref.rollout.n) // 8
    databuffer_number = min(databuffer_number, siirl_args.trainer.nnodes)
    return databuffer_number


@ray.remote(num_cpus=MAIN_RUNNER_CPU_RESERVATION)
class MainRunner:
    """
    A Ray actor responsible for orchestrating the entire RL training workflow.

    This actor handles loading configurations, scheduling task graphs, initializing
    process groups, and launching the distributed Ray trainers. Isolating this
    orchestration logic in a dedicated actor ensures the main process remains clean
    and that the setup process is managed within the Ray cluster.
    """

    def run(self, siirl_args: SiiRLArguments) -> None:
        """
        Executes the main training workflow.

        Args:
            siirl_args: A SiiRLArguments object containing all parsed configurations.
        """
        set_basic_config()
        from loguru import logger

        logger.info("MainRunner started. Beginning workflow setup...")
        start_time = time.time()

        # 1. Load environment configurations
        environment_handlers = None
        if siirl_args.dag.env_enable:
            assert siirl_args.dag.environment_path is not None, "Environment path must be provided when env_enable is True."
            logger.info(f"Loading environments from: {siirl_args.dag.environment_path}")
            environment_handlers = EnvironmentConfigLoader().load_environments_from_file(siirl_args.dag.environment_path)

        # 2. Init DataBuffer
        logger.info(f"Init DataBuffer with sharding number: {siirl_args.trainer.nnodes}")
        databuffer_number = get_databuffer_shard_number(siirl_args)
        data_buffer_handlers = init_data_buffer(databuffer_number)

        # 3. Load and configure the workerflow task graph (DAG)
        if siirl_args.dag.workflow_path is None:
            # If no workerflow path is provided, determine the default workflow config
            workflow_path = determine_workflow_config(self, siirl_args)
            logger.info(f"No workerflow path provided. Using {workflow_path} determined by adv_estimator: {siirl_args.algorithm.adv_estimator}")
        else:
            workflow_path = siirl_args.dag.workflow_path
        logger.info(f"Loading workerflow from: {siirl_args.dag.workflow_path}")
        if siirl_args.algorithm.adv_estimator == AdvantageEstimator.CPGD:
            siirl_args.actor_rollout_ref.actor.use_cpgd_loss = True
        workerflow_taskgraph = DAGConfigLoader.load_from_file(workflow_path)
        update_task_graph_node_configs(workerflow_taskgraph, siirl_args)
        display_node_config(workerflow_taskgraph)

        # 4. Schedule the task graph across available resources
        logger.info("Scheduling tasks across nodes and GPUs...")
        total_workers = siirl_args.trainer.nnodes * siirl_args.trainer.n_gpus_per_node
        task_scheduler = TaskScheduler(siirl_args.trainer.nnodes, siirl_args.trainer.n_gpus_per_node)
        rank_taskgraph_mapping = task_scheduler.schedule_and_assign_tasks([workerflow_taskgraph])
        log_schedule_assignments(rank_taskgraph_mapping, total_workers)
        unique_graphs_map = task_scheduler.get_unique_assigned_task_graphs()

        # 5. Create and configure process groups for communication
        logger.info("Initializing process groups for distributed communication...")
        process_group_manager = ProcessGroupManager(total_workers, rank_taskgraph_mapping)
        log_process_group_manager_details(process_group_manager, log_level="debug")
        # set process_group info into env for inference_actor
        inference_process_group = []
        inference_groups = process_group_manager.node_type_process_group_mapping["MODEL_INFERENCE"]
        for group_name in inference_groups:
            inference_process_group.append(process_group_manager.process_group_spec[group_name])
        os.environ["DGA_PROCESS_GROUP"] = str(inference_process_group)
        # 6. Initialize the main trainer
        logger.info("Initializing RayTrainer...")
        trainer = RayTrainer(
            config=siirl_args,
            process_group_manager=process_group_manager,
            rank_taskgraph_mapping=rank_taskgraph_mapping,
            unique_graphs_map=unique_graphs_map,
            data_buffer_handles=data_buffer_handlers,  # Placeholder for DataBuffer
            environments_handles=environment_handlers,
            device_name=siirl_args.trainer.device,
        )

        # 7. Initialize and start DAGWorkers
        logger.info("Initializing and starting DAG workers...")
        trainer.init_workers()
        trainer.start_workers()

        setup_duration = time.time() - start_time
        logger.info(f"Workflow setup and worker launch complete. Time cost: {setup_duration:.2f}s")


@hydra.main(config_path="config", config_name="ppo_dag_trainer", version_base=None)
def main(siirl_config: DictConfig) -> None:
    """
    Main entry point for launching the PPO DAG training job.

    This function initializes Ray, parses configurations using Hydra, and
    starts the MainRunner actor to orchestrate the distributed training workflow.

    Args:
        siirl_config: The configuration object provided by Hydra.
    """
    start_time = time.time()

    # Initialize Ray cluster if not already running
    if not ray.is_initialized():
        logger.info("Initializing local Ray cluster...")
        ray.init(runtime_env={"env_vars": RAY_RUNTIME_ENV_VARS}, num_cpus=siirl_config.ray_init.num_cpus)
    logger.success(f"Ray is initialized. Time cost: {(time.time() - start_time) * 1000:.2f} ms")

    # Parse the complete configuration into a structured object
    siirl_args = parse_config(siirl_config)
    log_dict_formatted(siirl_args.to_dict(), "SiiRLArguments")

    # Launch the main orchestration actor and wait for it to complete.
    logger.info("Starting MainRunner actor to orchestrate the job.")
    runner = MainRunner.remote()
    # This is a blocking call that waits for the remote `run` method to finish.
    ray.get(runner.run.remote(siirl_args))

    logger.success("MainRunner has completed its execution. Shutting down.")


if __name__ == "__main__":
    main()



# Copyright 2025, Shanghai Innovation Institute. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Manages a group of Ray actors for distributed workloads, handling their
creation, lifecycle, and communication.
"""

import os
import re
import time
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple
import asyncio

import ray
from loguru import logger
from ray.actor import ActorHandle
from ray.experimental.state.api import get_actor
from ray.util import list_named_actors

from siirl.utils.extras.device import get_device_name
from siirl.scheduler.process_group_manager import ProcessGroupManager
from siirl.utils.params import SiiRLArguments
from siirl.workers.base_worker import RayClassWithInitArgs, RayResourcePool, WorkerGroup, get_random_string, sort_placement_group_by_node_ip
from siirl.workers.dag import TaskGraph
from siirl.workers.dag_worker.dagworker import DAGWorker
from siirl.multiturn.agent_loop import AgentLoopManager


class DistributedEnv(Enum):
    """Enumeration for distributed environment variable keys."""

    MASTER_ADDR = "MASTER_ADDR"
    MASTER_PORT = "MASTER_PORT"
    WORLD_SIZE = "WORLD_SIZE"
    RANK = "RANK"
    WG_PREFIX = "WG_PREFIX"
    WG_BACKEND = "WG_BACKEND"
    RAY_LOCAL_WORLD_SIZE = "RAY_LOCAL_WORLD_SIZE"
    RAY_LOCAL_RANK = "RAY_LOCAL_RANK"
    DGA_PROCESS_GROUP = "DGA_PROCESS_GROUP"


# --- Constants ---
ACTOR_STATE_ALIVE = "ALIVE"
REGISTER_CENTER_POLL_INTERVAL_S = 1
REGISTER_CENTER_LOG_INTERVAL_S = 30
RAY_BACKEND = "ray"


class RayActorManager(WorkerGroup):
    """
    Manages the lifecycle of a group of distributed Ray actors (workers).

    This class handles the creation of actors based on resource availability,
    assigns ranks, and sets up the necessary environment variables for
    distributed communication. It provides a unified interface to execute
    methods synchronously or asynchronously across all or specific workers.

    Attributes:
        worker_names (List[str]): A list of the generated names for each actor.
        master_address (str): The network address of the rank 0 worker.
        master_port (str): The network port of the rank 0 worker.
        workers (List[ActorHandle]): The list of Ray actor handles managed by this group.
        world_size (int): The total number of workers in the group.
    """

    def __init__(
        self,
        resource_pool: RayResourcePool,
        base_config: SiiRLArguments,
        process_manager: ProcessGroupManager,
        rank_taskgraph_mapping: Dict[int, "TaskGraph"],
        data_buffer_handles: List[ActorHandle],
        environments_handles: Optional[Dict[int, ActorHandle]] = None,
        bin_pack: bool = True,
        name_prefix: Optional[str] = None,
        ray_wait_register_center_timeout: int = 300,
        device_name="cuda",
        **kwargs,
    ) -> None:
        """
        Initializes the RayActorManager.

        Args:
            resource_pool: The pool of resources available for placing actors.
            base_config: Base configuration arguments for the workers.
            process_manager: Manager for the distributed process group.
            rank_taskgraph_mapping: Mapping of worker ranks to their task graphs.
            data_buffer_handles: List of handles to shared data buffers.
            environments_handles: Optional mapping of ranks to environment handles.
            bin_pack: If True, use strict packing strategy for placement groups.
            name_prefix: A custom prefix for actor names. A random one is
                         generated if None.
            ray_wait_register_center_timeout: Seconds to wait for the rank 0
                                              registration actor to appear.
            **kwargs: Additional arguments for the base WorkerGroup.
        """
        super().__init__(resource_pool=resource_pool, **kwargs)

        self.name_prefix: str = get_random_string(length=6) if name_prefix is None else name_prefix
        self._ray_wait_register_center_timeout = ray_wait_register_center_timeout

        self._worker_names: List[str] = []
        self._world_size: int = resource_pool.world_size
        self._master_addr: Optional[str] = None
        self._master_port: Optional[str] = None

        self.base_config = base_config
        self.process_manager = process_manager
        self.rank_taskgraph_mapping = rank_taskgraph_mapping
        self.data_buffer_handles = data_buffer_handles
        self.environments_handles = environments_handles
        self.device_name = device_name

        # Prepare the Ray actor class with its initial arguments.
        self.ray_actor_class = RayClassWithInitArgs(
            ray.remote(DAGWorker),
            config=self.base_config,
            process_group_manager=self.process_manager,
            taskgraph_mapping=self.rank_taskgraph_mapping,
            data_buffers=self.data_buffer_handles,
            environments=self.environments_handles,
            device_name=self.device_name,
        )

        self._initialize_workers(resource_pool=resource_pool, bin_pack=bin_pack)

    def _initialize_workers(self, resource_pool: RayResourcePool, bin_pack: bool) -> None:
        """
        Creates and configures all worker actors based on the resource pool.

        This method orchestrates the creation of placement groups, iterates through
        them to launch actors with the correct rank and environment variables,
        and establishes the master address for distributed coordination.
        """
        strategy = "STRICT_PACK" if bin_pack else "PACK"
        placement_groups = resource_pool.get_placement_groups(strategy=strategy, device_name=self.device_name)
        sorted_pgs = sort_placement_group_by_node_ip(placement_groups)

        num_gpus_per_worker = 1 / resource_pool.max_colocate_count
        local_world_size = resource_pool.store[0]
        rank = -1

        for pg_index, placement_group in enumerate(sorted_pgs):
            if local_world_size > placement_group.bundle_count:
                raise ValueError(f"Placement group for '{self.name_prefix}' has too few bundles ({placement_group.bundle_count}) to support the required local world size ({local_world_size}).")

            for local_rank in range(local_world_size):
                rank += 1
                worker = self._create_worker_actor(
                    rank=rank,
                    local_rank=local_rank,
                    local_world_size=local_world_size,
                    pg_index=pg_index,
                    placement_group=placement_group,
                    num_gpus_per_worker=num_gpus_per_worker,
                    use_gpu=resource_pool.use_gpu,
                    device_name=self.device_name,
                )
                self._workers.append(worker)

                if rank == 0:
                    # Rank 0 worker is special: it establishes the master
                    # address and port for the entire worker group.
                    self._master_addr, self._master_port = self._get_register_center_and_master_info()
        work_futures = self.map_async(method_name="init_graph")
        ray.get(work_futures)
        # only support single agent
        if self.base_config.actor_rollout_ref.rollout.mode == 'async':
            tp_size = self.base_config.actor_rollout_ref.rollout.tensor_model_parallel_size
            world_size = len(self._workers)
            dp_size = world_size // tp_size
            self.async_rollout_manager = []
            futures = self._init_async_rollout_manaager(dp_size)
            
            loop = asyncio.get_event_loop()
            self.async_rollout_manager = loop.run_until_complete(futures)
            for manager in self.async_rollout_manager:
                manager.sleep()
            for i in range(len(self._workers)):
                if i % tp_size == 0:
                    ray.get(self._workers[i].set_async_rollout_manager.remote(self.async_rollout_manager[i // tp_size]))
    async def _init_async_rollout_manaager(self,dp_size):
         # add sync_rollout manager
        futures = []
        for dp_rank in range(dp_size):
            futures.append(self._create_async_rollout_manaager(dp_size, dp_rank))
        return await asyncio.gather(*futures)
    async def _create_async_rollout_manaager(self, dp_size, dp_rank):
        agent_manager = AgentLoopManager(self.base_config.actor_rollout_ref, dp_size, dp_rank, self.name_prefix)
        await agent_manager.init_model()
        return agent_manager
        
    def _get_register_center_and_master_info(self) -> Tuple[str, str]:
        """
        Waits for the registration actor to be available and fetches the
        master address and port from it.

        Returns:
            A tuple containing the master address and master port.

        Raises:
            TimeoutError: If the registration actor cannot be found within the
                          configured timeout.
        """
        register_center_name = f"{self.name_prefix}_register_center"
        logger.info(f"Waiting for registration center actor: '{register_center_name}'...")
        start_time = time.time()

        while time.time() - start_time < self._ray_wait_register_center_timeout:
            try:
                # Use list_named_actors for a more robust check.
                if register_center_name in list_named_actors():
                    register_center_actor = ray.get_actor(register_center_name)
                    logger.success(f"Successfully connected to '{register_center_name}'.")
                    rank_zero_info = ray.get(register_center_actor.get_rank_zero_info.remote())

                    master_addr = rank_zero_info[DistributedEnv.MASTER_ADDR.value]
                    master_port = rank_zero_info[DistributedEnv.MASTER_PORT.value]
                    return master_addr, master_port
            except Exception as e:
                logger.warning(f"Attempt to get register center failed, will retry. Error: {e}")

            elapsed = int(time.time() - start_time)
            if elapsed > 0 and elapsed % REGISTER_CENTER_LOG_INTERVAL_S == 0:
                logger.warning(f"Still waiting for '{register_center_name}'. Elapsed: {elapsed}s / {self._ray_wait_register_center_timeout}s.")
            time.sleep(REGISTER_CENTER_POLL_INTERVAL_S)

        raise TimeoutError(
            f"Failed to get register_center_actor '{register_center_name}' "
            f"within {self._ray_wait_register_center_timeout} seconds. "
            f"Existing actors: {list_named_actors(all_namespaces=True)}. "
            "Ensure Ray resources from previous runs are cleaned up or "
            "increase 'trainer.ray_wait_register_center_timeout'."
        )

    def _create_worker_actor(
        self,
        rank: int,
        local_rank: int,
        local_world_size: int,
        pg_index: int,
        placement_group: "ray.util.placement_group.PlacementGroup",
        num_gpus_per_worker: float,
        use_gpu: bool,
        device_name: str,
    ) -> ActorHandle:
        """
        Configures and creates a single worker actor.

        Args:
            rank: The global rank of the worker.
            local_rank: The rank of the worker on its local node.
            local_world_size: The number of workers on the local node.
            pg_index: The index of the placement group being used.
            placement_group: The Ray placement group for this worker.
            num_gpus_per_worker: The number of GPUs to assign to the worker.
            use_gpu: A boolean indicating if the worker should use a GPU.
            device_name: The name of the device to use ("cuda" or "npu").

        Returns:
            The handle to the newly created Ray actor.
        """
        # --- 1. Prepare Environment Variables ---
        env_vars = {
            DistributedEnv.WORLD_SIZE.value: str(self.world_size),
            DistributedEnv.RANK.value: str(rank),
            DistributedEnv.WG_PREFIX.value: self.name_prefix,
            DistributedEnv.WG_BACKEND.value: RAY_BACKEND,
            DistributedEnv.RAY_LOCAL_WORLD_SIZE.value: str(local_world_size),
            DistributedEnv.RAY_LOCAL_RANK.value: str(local_rank),
            DistributedEnv.DGA_PROCESS_GROUP.value: os.environ.get("DGA_PROCESS_GROUP", ""),
        }
        if rank != 0:
            if not self._master_addr or not self._master_port:
                raise ConnectionError("Master address and port not set before creating non-zero rank workers.")
            env_vars[DistributedEnv.MASTER_ADDR.value] = self._master_addr
            env_vars[DistributedEnv.MASTER_PORT.value] = self._master_port

        # --- 2. Generate a unique and descriptive actor name ---
        base_class_repr = type(self.ray_actor_class.cls).__name__  # e.g., "ActorClass(DAGWorker)"
        match = re.search(r"ActorClass\(([^)]+)\)", base_class_repr)
        actor_class_name = match.group(1) if match else base_class_repr
        actor_name = f"{self.name_prefix}_{actor_class_name}_{pg_index}:{local_rank}"
        self._worker_names.append(actor_name)

        # --- 3. Set actor-specific options ---
        self.ray_actor_class.update_options(
            {
                "runtime_env": {"env_vars": env_vars},
                "name": actor_name,
            }
        )

        # --- 4. Create the actor ---
        logger.debug(f"Creating actor '{actor_name}' with rank {rank}.")
        worker = self.ray_actor_class(placement_group=placement_group, placement_group_bundle_idx=local_rank, use_gpu=use_gpu, num_gpus=num_gpus_per_worker, device_name=device_name)
        return worker

    def _is_worker_alive(self, worker: ActorHandle) -> bool:
        """
        Checks if a given worker actor is in the 'ALIVE' state.

        Note: This uses a Ray experimental API, which may change in the future.

        Args:
            worker: The actor handle to check.

        Returns:
            True if the worker is alive, False otherwise.
        """
        try:
            worker_state_dict = get_actor(worker._actor_id.hex())
            return worker_state_dict.get("state", "undefined") == ACTOR_STATE_ALIVE if worker_state_dict else False
        except Exception:
            return False

    def _invoke_on_worker(self, worker: ActorHandle, method_name: str, *args: Any, **kwargs: Any) -> Any:
        """Invokes a method on a single remote worker actor."""
        remote_method = getattr(worker, method_name)
        return remote_method.remote(*args, **kwargs)

    def map_sync(self, method_name: str, *args: Any, **kwargs: Any) -> List[Any]:
        """
        Executes a method on all workers and waits for all results to complete.

        This is the synchronous (blocking) version of `map_async`.

        Args:
            method_name: The name of the method to execute.
            *args: Positional arguments for the method.
            **kwargs: Keyword arguments for the method.

        Returns:
            A list containing the results from each worker.
        """
        return ray.get(self.map_async(method_name, *args, **kwargs))

    def map_async(self, method_name: str, *args: Any, **kwargs: Any) -> List[ray.ObjectRef]:
        """
        Executes a method on all workers asynchronously.

        Special Behavior:
        If all positional and keyword arguments are lists of the same length
        as the number of workers, the arguments are "unzipped" and distributed.
        The i-th worker receives the i-th element from each argument list.
        Otherwise, all workers receive the same arguments.

        Args:
            method_name: The name of the method to execute.
            *args: Positional arguments for the method.
            **kwargs: Keyword arguments for the method.

        Returns:
            A list of ray.ObjectRef handles to the results.
        """
        num_workers = len(self._workers)

        # Check for the special argument-splitting case
        all_args_are_distributable = all(isinstance(arg, list) and len(arg) == num_workers for arg in args) and all(isinstance(val, list) and len(val) == num_workers for val in kwargs.values())

        if all_args_are_distributable and (args or kwargs):
            futures = []
            for i in range(num_workers):
                # Slice arguments for the i-th worker
                sliced_args = tuple(arg[i] for arg in args)
                sliced_kwargs = {k: v[i] for k, v in kwargs.items()}
                futures.append(self._invoke_on_worker(self._workers[i], method_name, *sliced_args, **sliced_kwargs))
            return futures

        # Default case: all workers get the same arguments
        return [self._invoke_on_worker(w, method_name, *args, **kwargs) for w in self._workers]

    def map(self, method_name: str, *args: Any, **kwargs: Any) -> List[ray.ObjectRef]:
        """
        Alias for `map_async`. Executes a method on all workers asynchronously.
        """
        return self.map_async(method_name, *args, **kwargs)

    @property
    def worker_names(self) -> List[str]:
        """Returns the names of all managed workers."""
        return self._worker_names

    @property
    def master_address(self) -> Optional[str]:
        """Returns the address of the master (rank 0) worker."""
        return self._master_addr

    @property
    def master_port(self) -> Optional[str]:
        """Returns the port of the master (rank 0) worker."""
        return self._master_port

    @property
    def workers(self) -> List[ActorHandle]:
        """Returns the list of Ray actor handles."""
        return self._workers

    @property
    def world_size(self) -> int:
        """Returns the total number of workers."""
        return self._world_size



# Copyright 2024 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import random
from typing import Optional

import numpy as np
import torch
import torch.distributed
from megatron.core import mpu, tensor_parallel
from megatron.core.dist_checkpointing.mapping import ShardedObject
from transformers import GenerationConfig

from siirl.models.weight_loader_registry import get_weight_saver
from siirl.utils.extras.device import is_cuda_available, is_npu_available
from siirl.utils.extras.fs import is_non_local
from siirl.utils.megatron.megatron_utils import (
    get_hf_config_and_tokenizer_checkpoint_path,
    get_hf_model_checkpoint_path,
    get_model_checkpoint_path,
    get_optimizer_checkpoint_path,
    get_rng_states_checkpoint_path,
)

from .checkpoint_manager import BaseCheckpointManager


class MegatronCheckpointManager(BaseCheckpointManager):
    """
    A checkpoint manager that saves and loads
    - model
    - optimizer
    - lr_scheduler
    - extra_states
    in a SPMD way.

    We save
    - sharded model states and optimizer states
    - full lr_scheduler states
    - huggingface tokenizer/processor and config for ckpt merge
    """

    def __init__(
        self,
        config,
        model_config,
        role,
        model: torch.nn.ModuleList,
        arch: str,
        hf_config,
        param_dtype: torch.dtype,
        share_embeddings_and_output_weights: bool,
        tokenizer,
        optimizer,
        use_distributed_optimizer: bool,
        checkpoint_contents: Optional[list] = None,
        **kwargs,
    ):
        if checkpoint_contents is None:
            checkpoint_contents = ["model", "optimizer", "extra"]
        super().__init__(
            model,
            optimizer=optimizer,
            lr_scheduler=None,
            processing_class=tokenizer,
            checkpoint_contents=checkpoint_contents,
        )
        self.arch = arch
        self.config = config
        self.role = role
        self.is_value_model = False
        if self.role in ["reward", "critic"]:
            self.is_value_model = True
        self.model_config = model_config
        self.hf_config = hf_config
        self.param_dtype = param_dtype
        self.share_embeddings_and_output_weights = share_embeddings_and_output_weights
        self.model_path = self.config.model.path
        self.use_distributed_optimizer = use_distributed_optimizer

        self.rank = torch.distributed.get_rank()

        self.weight_saver = get_weight_saver(self.arch)

    def get_rng_state(self, use_dist_ckpt: bool = False, data_parallel_random_init: bool = False):
        """collect rng state across data parallel ranks"""
        rng_state = {
            "random_rng_state": random.getstate(),
            "np_rng_state": np.random.get_state(),
            "torch_rng_state": torch.get_rng_state(),
            "rng_tracker_states": tensor_parallel.get_cuda_rng_tracker().get_states(),
        }

        if is_cuda_available:
            rng_state["cuda_rng_state"] = torch.cuda.get_rng_state()
        elif is_npu_available:
            rng_state["npu_rng_state"] = torch.npu.get_rng_state()

        rng_state_list = None
        if torch.distributed.is_initialized() and mpu.get_data_parallel_world_size() > 1 and data_parallel_random_init:
            rng_state_list = [None for i in range(mpu.get_data_parallel_world_size())]
            torch.distributed.all_gather_object(rng_state_list, rng_state, group=mpu.get_data_parallel_group())
        else:
            rng_state_list = [rng_state]

        if use_dist_ckpt:
            pp_rank = mpu.get_pipeline_model_parallel_rank()
            pp_size = mpu.get_pipeline_model_parallel_world_size()
            tp_rank = mpu.get_tensor_model_parallel_rank()
            tp_size = mpu.get_tensor_model_parallel_world_size()
            cp_rank = mpu.get_context_parallel_rank()
            cp_size = mpu.get_context_parallel_world_size()
            rng_state_list = ShardedObject(
                "rng_state",
                rng_state_list,
                (pp_size, tp_size, cp_size),
                (pp_rank, tp_rank, cp_rank),
                replica_id=mpu.get_data_parallel_rank(with_context_parallel=True),
            )

        return rng_state_list

    def get_checkpoint_name(
        self,
        checkpoints_path,
        pipeline_parallel=None,
        tensor_rank=None,
        pipeline_rank=None,
        cp_rank=None,
        expert_parallel=None,
        expert_rank=None,
        return_base_dir=True,
        basename="model.pt",
    ):
        """Determine the directory name for this rank's checkpoint."""
        # Use both the tensor and pipeline MP rank.
        if pipeline_parallel is None:
            pipeline_parallel = mpu.get_pipeline_model_parallel_world_size() > 1
        if tensor_rank is None:
            tensor_rank = mpu.get_tensor_model_parallel_rank()
        if pipeline_rank is None:
            pipeline_rank = mpu.get_pipeline_model_parallel_rank()
        if cp_rank is None:
            cp_rank = mpu.get_context_parallel_rank()
        if expert_parallel is None:
            expert_parallel = mpu.get_expert_model_parallel_world_size() > 1
        if expert_rank is None:
            expert_rank = mpu.get_expert_model_parallel_rank()

        # Use both the tensor and pipeline MP rank. If using the distributed
        # optimizer, then the optimizer's path must additionally include the
        # data parallel rank.

        # due to the fact that models are identical across cp ranks, cp rank is not used in the checkpoint path
        if not pipeline_parallel:
            common_path = os.path.join(checkpoints_path, f"mp_rank_{tensor_rank:02d}")
        else:
            common_path = os.path.join(checkpoints_path, f"mp_rank_{tensor_rank:02d}_{pipeline_rank:03d}")

        if expert_parallel:
            common_path = common_path + f"_{expert_rank:03d}"

        os.makedirs(common_path, exist_ok=True)

        if return_base_dir:
            return common_path
        return os.path.join(common_path, basename)

    def load_optimizer(self, ckpt_path):
        # TODO: Check Optimizer format and distributed optimizer
        optimizer_path = get_optimizer_checkpoint_path(ckpt_path)
        print(f"Loading optimizer from {optimizer_path}")
        self.optimizer.load_parameter_state(optimizer_path)

    def load_rng_states(self, ckpt_path, data_parallel_random_init=False, use_dist_ckpt=False):
        rng_state_path = get_rng_states_checkpoint_path(ckpt_path, only_rank0_save=False)
        print(f"Loading rng states from {rng_state_path}")
        rng_state = torch.load(rng_state_path, weights_only=False)
        # access rng_state for data parallel rank
        if not use_dist_ckpt:
            rng_state = rng_state[mpu.get_data_parallel_rank()] if data_parallel_random_init else rng_state[0]
        random.setstate(rng_state["random_rng_state"])
        np.random.set_state(rng_state["np_rng_state"])
        torch.set_rng_state(rng_state["torch_rng_state"])
        
        if is_cuda_available:
            torch.cuda.set_rng_state(rng_state["cuda_rng_state"])
        elif is_npu_available:
            torch.npu.set_rng_state(rng_state["npu_rng_state"])
        
        # Check for empty states array
        if not rng_state["rng_tracker_states"]:
            raise KeyError
        tensor_parallel.get_cuda_rng_tracker().set_states(rng_state["rng_tracker_states"])

    def load_checkpoint(self, local_path: str, hdfs_path: str = None, del_local_after_load=False):
        if local_path is None:
            return

        if "model" in self.checkpoint_contents:
            model_path = get_model_checkpoint_path(local_path)
            ckpt_name = self.get_checkpoint_name(model_path, return_base_dir=False)
            state_dicts = torch.load(os.path.join(ckpt_name), weights_only=False)
            assert len(state_dicts) == len(self.model), f"state_dicts length: {len(state_dicts)} mismatch with model length: {len(self.model)}"
            for vpp_rank, (state_dict, model) in enumerate(zip(state_dicts, self.model)):
                model.load_state_dict(state_dict)
            print(f"Loaded sharded model checkpoint from {model_path}")

        if "optimizer" in self.checkpoint_contents:
            self.load_optimizer(local_path)

        if "extra" in self.checkpoint_contents:
            self.load_rng_states(local_path)

        if del_local_after_load:
            try:
                os.remove(local_path) if is_non_local(local_path) else None
            except Exception as e:
                print(f"[rank-{self.rank}]: remove local resume ckpt file after loading failed, exception {e} will be ignored")

    def save_checkpoint(self, local_path: str, hdfs_path: str = None, global_step: int = 0, max_ckpt_to_keep=None):
        # record the previous global step
        self.previous_global_step = global_step

        # remove previous local_path
        if max_ckpt_to_keep and isinstance(max_ckpt_to_keep, int) and max_ckpt_to_keep > 0 and len(self.previous_saved_paths) >= max_ckpt_to_keep:
            keep_start = len(self.previous_saved_paths) - max_ckpt_to_keep + 1
            self.remove_previous_save_local_path(self.previous_saved_paths[:keep_start])
            self.previous_saved_paths = self.previous_saved_paths[keep_start:]

        local_path = self.local_mkdir(local_path)

        # Save Model
        if "model" in self.checkpoint_contents and mpu.get_data_parallel_rank() == 0:
            state_dicts = []

            for vpp_rank, model in enumerate(self.model):
                state_dict = model.state_dict()
                state_dicts.append(state_dict)

            print(f"Saving sharded model checkpoint to {local_path}")
            model_ckpt_path = get_model_checkpoint_path(local_path)
            hf_config_and_tokenizer_path = get_hf_config_and_tokenizer_checkpoint_path(local_path)
            ckpt_name = self.get_checkpoint_name(model_ckpt_path, return_base_dir=False)
            torch.save(state_dicts, os.path.join(ckpt_name))

            print(f"Saved checkpoint to {model_ckpt_path}")
            if self.rank == 0:
                self.processing_class.save_pretrained(hf_config_and_tokenizer_path)
                self.hf_config.save_pretrained(hf_config_and_tokenizer_path)
                if hasattr(self.hf_config, "name_or_path") and self.hf_config.name_or_path:
                    try:
                        generation_config = GenerationConfig.from_pretrained(self.hf_config.name_or_path)
                        generation_config.save_pretrained(hf_config_and_tokenizer_path)
                    except Exception:
                        # if the generation config isn't available, we don't save it
                        pass
                if hdfs_path is not None:
                    print(f"Uploading checkpoint to {hdfs_path}")
                    from siirl.utils.extras.hdfs_io import hdfs_io

                    hdfs_io.makedirs(hdfs_path, exist_ok=True)
                    hdfs_io.copy(src=model_ckpt_path, dst=hdfs_path, dirs_exist_ok=True)
                    hdfs_io.copy(src=hf_config_and_tokenizer_path, dst=hdfs_path, dirs_exist_ok=True)

        if "hf_model" in self.checkpoint_contents:
            # wait for everyone to dump to local
            state_dict = self.weight_saver(
                self.model,
                self.hf_config,
                dtype=self.param_dtype,
                is_value_model=self.is_value_model,
                tie_word_embeddings=self.share_embeddings_and_output_weights,
            )

            torch.distributed.barrier()
            if self.rank == 0:
                print(f"self.param_dtype: {self.param_dtype}")
                for key in state_dict.keys():
                    print(f"state_dict[key].dtype: {key} {state_dict[key].dtype}")
                hf_model_ckpt_path = get_hf_model_checkpoint_path(local_path)
                import warnings

                from accelerate import init_empty_weights

                with init_empty_weights(), warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    if "mistral7b-rm" in self.config.model.path:
                        from transformers import MistralForSequenceClassification

                        model = MistralForSequenceClassification.from_pretrained(self.config.model.path)  # use score head instead of lm_head
                        state_dict["score.weight"] = state_dict["score.weight"]
                    else:
                        from transformers import AutoModelForCausalLM

                        model = AutoModelForCausalLM.from_pretrained(self.config.model.path, torch_dtype="auto")
                model.save_pretrained(hf_model_ckpt_path, state_dict=state_dict)
                self.processing_class.save_pretrained(hf_model_ckpt_path)

                if hdfs_path is not None:
                    print(f"Uploading checkpoint to {hdfs_path}")
                    from siirl.utils.extras.hdfs_io import hdfs_io

                    hdfs_io.makedirs(hdfs_path, exist_ok=True)
                    hdfs_io.copy(src=hf_model_ckpt_path, dst=hdfs_path, dirs_exist_ok=True)

        # Save Optimizer
        if "optimizer" in self.checkpoint_contents:
            torch.distributed.barrier()

            optimizer_path = get_optimizer_checkpoint_path(local_path)
            self.optimizer.save_parameter_state(optimizer_path)
            if self.rank == 0:
                print(f"saving optimizer state to {optimizer_path}")

        # Save RNG States
        if "extra" in self.checkpoint_contents:
            torch.distributed.barrier()

            rng_state_path = get_rng_states_checkpoint_path(local_path, only_rank0_save=False)
            rng_state = self.get_rng_state()
            torch.save(rng_state, rng_state_path)
            print(f"Rank {self.rank} saving rng states to {rng_state_path}")

        self.previous_saved_paths.append(local_path)



"""
This logic is largely copied from the Hendrycks' MATH release (math_equivalence), and borrowed from:
- https://github.com/microsoft/ProphetNet/tree/master/CRITIC
- https://github.com/openai/prm800k
- https://github.com/microsoft/ToRA/blob/main/src/eval/grader.py
- https://github.com/deepseek-ai/DeepSeek-Math/blob/main/evaluation/eval/eval_utils.py
"""

import re
import regex
import multiprocessing
from math import isclose
from typing import Union

from sympy import simplify, N
from sympy.parsing.sympy_parser import parse_expr
from sympy.parsing.latex import parse_latex
from latex2sympy2 import latex2sympy

# from .parser import choice_answer_clean, strip_string
# from parser import choice_answer_clean


def choice_answer_clean(pred: str):
    pred = pred.strip("\n").rstrip(".").rstrip("/").strip(" ").lstrip(":")
    # Clean the answer based on the dataset
    tmp = re.findall(r"\b(A|B|C|D|E)\b", pred.upper())
    if tmp:
        pred = tmp
    else:
        pred = [pred.strip().strip(".")]
    pred = pred[-1]
    # Remove the period at the end, again!
    pred = pred.rstrip(".").rstrip("/")
    return pred


def parse_digits(num):
    num = regex.sub(",", "", str(num))
    try:
        return float(num)
    except:
        if num.endswith("%"):
            num = num[:-1]
            if num.endswith("\\"):
                num = num[:-1]
            try:
                return float(num) / 100
            except:
                pass
    return None


def is_digit(num):
    # paired with parse_digits
    return parse_digits(num) is not None


def str_to_pmatrix(input_str):
    input_str = input_str.strip()
    matrix_str = re.findall(r"\{.*,.*\}", input_str)
    pmatrix_list = []

    for m in matrix_str:
        m = m.strip("{}")
        pmatrix = r"\begin{pmatrix}" + m.replace(",", "\\") + r"\end{pmatrix}"
        pmatrix_list.append(pmatrix)

    return ", ".join(pmatrix_list)


def math_equal(
    prediction: Union[bool, float, str],
    reference: Union[float, str],
    include_percentage: bool = True,
    is_close: bool = True,
    timeout: bool = False,
) -> bool:
    """
    Exact match of math if and only if:
    1. numerical equal: both can convert to float and are equal
    2. symbolic equal: both can convert to sympy expression and are equal
    """
    # print("Judge:", prediction, reference)
    if prediction is None or reference is None:
        return False
    if str(prediction.strip().lower()) == str(reference.strip().lower()):
        return True
    if reference in ["A", "B", "C", "D", "E"] and choice_answer_clean(prediction) == reference:
        return True

    try:  # 1. numerical equal
        if is_digit(prediction) and is_digit(reference):
            prediction = parse_digits(prediction)
            reference = parse_digits(reference)
            # number questions
            if include_percentage:
                gt_result = [reference / 100, reference, reference * 100]
            else:
                gt_result = [reference]
            for item in gt_result:
                try:
                    if is_close:
                        if numeric_equal(prediction, item):
                            return True
                    else:
                        if item == prediction:
                            return True
                except Exception:
                    continue
            return False
    except:
        pass

    if not prediction and prediction not in [0, False]:
        return False

    # 2. symbolic equal
    reference = str(reference).strip()
    prediction = str(prediction).strip()

    ## pmatrix (amps)
    if "pmatrix" in prediction and not "pmatrix" in reference:
        reference = str_to_pmatrix(reference)

    ## deal with [], (), {}
    pred_str, ref_str = prediction, reference
    if (prediction.startswith("[") and prediction.endswith("]") and not reference.startswith("(")) or (prediction.startswith("(") and prediction.endswith(")") and not reference.startswith("[")):
        pred_str = pred_str.strip("[]()")
        ref_str = ref_str.strip("[]()")
    for s in ["{", "}", "(", ")"]:
        ref_str = ref_str.replace(s, "")
        pred_str = pred_str.replace(s, "")
    if pred_str.lower() == ref_str.lower():
        return True

    ## [a, b] vs. [c, d], return a==c and b==d
    if regex.match(r"(\(|\[).+(\)|\])", prediction) is not None and regex.match(r"(\(|\[).+(\)|\])", reference) is not None:
        pred_parts = prediction[1:-1].split(",")
        ref_parts = reference[1:-1].split(",")
        if len(pred_parts) == len(ref_parts):
            if all([math_equal(pred_parts[i], ref_parts[i], include_percentage, is_close) for i in range(len(pred_parts))]):
                return True
    if (
        (prediction.startswith("\\begin{pmatrix}") or prediction.startswith("\\begin{bmatrix}"))
        and (prediction.endswith("\\end{pmatrix}") or prediction.endswith("\\end{bmatrix}"))
        and (reference.startswith("\\begin{pmatrix}") or reference.startswith("\\begin{bmatrix}"))
        and (reference.endswith("\\end{pmatrix}") or reference.endswith("\\end{bmatrix}"))
    ):
        pred_lines = [line.strip() for line in prediction[len("\\begin{pmatrix}") : -len("\\end{pmatrix}")].split("\\\\") if line.strip()]
        ref_lines = [line.strip() for line in reference[len("\\begin{pmatrix}") : -len("\\end{pmatrix}")].split("\\\\") if line.strip()]
        matched = True
        if len(pred_lines) == len(ref_lines):
            for pred_line, ref_line in zip(pred_lines, ref_lines):
                pred_parts = pred_line.split("&")
                ref_parts = ref_line.split("&")
                if len(pred_parts) == len(ref_parts):
                    if not all(
                        [
                            math_equal(
                                pred_parts[i],
                                ref_parts[i],
                                include_percentage,
                                is_close,
                            )
                            for i in range(len(pred_parts))
                        ]
                    ):
                        matched = False
                        break
                else:
                    matched = False
                if not matched:
                    break
        else:
            matched = False
        if matched:
            return True

    if prediction.count("=") == 1 and reference.count("=") == 1:
        pred = prediction.split("=")
        pred = f"{pred[0].strip()} - ({pred[1].strip()})"
        ref = reference.split("=")
        ref = f"{ref[0].strip()} - ({ref[1].strip()})"
        if symbolic_equal(pred, ref) or symbolic_equal(f"-({pred})", ref):
            return True
    elif prediction.count("=") == 1 and len(prediction.split("=")[0].strip()) <= 2 and "=" not in reference:
        if math_equal(prediction.split("=")[1], reference, include_percentage, is_close):
            return True
    elif reference.count("=") == 1 and len(reference.split("=")[0].strip()) <= 2 and "=" not in prediction:
        if math_equal(prediction, reference.split("=")[1], include_percentage, is_close):
            return True

    # symbolic equal with sympy
    if timeout:
        if call_with_timeout(symbolic_equal_process, prediction, reference):
            return True
    else:
        if symbolic_equal(prediction, reference):
            return True

    return False


def math_equal_process(param):
    return math_equal(param[-2], param[-1])


def numeric_equal(prediction: float, reference: float):
    # Note that relative tolerance has significant impact
    # on the result of the synthesized GSM-Hard dataset
    # if reference.is_integer():
    #     return isclose(reference, round(prediction), abs_tol=1e-4)
    # else:
    # prediction = round(prediction, len(str(reference).split(".")[-1]))
    return isclose(reference, prediction, rel_tol=1e-4)


def symbolic_equal(a, b):
    def _parse(s):
        for f in [parse_latex, parse_expr, latex2sympy]:
            try:
                return f(s.replace("\\\\", "\\"))
            except:
                try:
                    return f(s)
                except:
                    pass
        return s

    a = _parse(a)
    b = _parse(b)

    # direct equal
    try:
        if str(a) == str(b) or a == b:
            return True
    except:
        pass

    # simplify equal
    try:
        if a.equals(b) or simplify(a - b) == 0:
            return True
    except:
        pass

    # equation equal
    try:
        if (abs(a.lhs - a.rhs)).equals(abs(b.lhs - b.rhs)):
            return True
    except:
        pass

    try:
        if numeric_equal(float(N(a)), float(N(b))):
            return True
    except:
        pass

    # matrix
    try:
        # if a and b are matrix
        if a.shape == b.shape:
            _a = a.applyfunc(lambda x: round(x, 3))
            _b = b.applyfunc(lambda x: round(x, 3))
            if _a.equals(_b):
                return True
    except:
        pass

    return False


def symbolic_equal_process(a, b, output_queue):
    result = symbolic_equal(a, b)
    output_queue.put(result)


def call_with_timeout(func, *args, timeout=1, **kwargs):
    output_queue = multiprocessing.Queue()
    process_args = args + (output_queue,)
    process = multiprocessing.Process(target=func, args=process_args, kwargs=kwargs)
    process.start()
    process.join(timeout)

    if process.is_alive():
        process.terminate()
        process.join()
        return False

    return output_queue.get()


def _test_math_equal():
    # print(math_equal("0.0833333333333333", "\\frac{1}{12}"))
    # print(math_equal("(1,4.5)", "(1,\\frac{9}{2})"))
    # print(math_equal("\\frac{x}{7}+\\frac{2}{7}", "\\frac{x+2}{7}", timeout=True))
    # print(math_equal("\\sec^2(y)", "\\tan^2(y)+1", timeout=True))
    # print(math_equal("\\begin{pmatrix}-\\frac{7}{4}&-2\\\\4&\\frac{1}{4}\\end{pmatrix}", "(\\begin{pmatrix}-\\frac{7}{4}&-2\\\\4&\\frac{1}{4}\\\\\\end{pmatrix})", timeout=True))

    # pred = '\\begin{pmatrix}\\frac{1}{3x^{2/3}}&0&0\\\\0&1&0\\\\-\\sin(x)&0&0\\end{pmatrix}'
    # gt = '(\\begin{pmatrix}\\frac{1}{3\\sqrt[3]{x}^2}&0&0\\\\0&1&0\\\\-\\sin(x)&0&0\\\\\\end{pmatrix})'

    # pred= '-\\frac{8x^2}{9(x^2-2)^{5/3}}+\\frac{2}{3(x^2-2)^{2/3}}'
    # gt= '-\\frac{2(x^2+6)}{9(x^2-2)\\sqrt[3]{x^2-2}^2}'

    # pred =  '-34x-45y+20z-100=0'
    # gt = '34x+45y-20z+100=0'

    # pred = '\\frac{100}{3}'
    # gt = '33.3'

    # pred = '\\begin{pmatrix}0.290243531202435\\\\0.196008371385084\\\\-0.186381278538813\\end{pmatrix}'
    # gt = '(\\begin{pmatrix}0.29\\\\0.196\\\\-0.186\\\\\\end{pmatrix})'

    # pred = '\\frac{\\sqrt{\\sqrt{11}+\\sqrt{194}}}{2\\sqrt{33}+15}'
    # gt = '\\frac{\\sqrt{\\sqrt{11}+\\sqrt{194}}}{15+2\\sqrt{33}}'

    # pred = '(+5)(b+2)'
    # gt = '(a+5)(b+2)'

    # pred = '\\frac{1+\\sqrt{5}}{2}'
    # gt = '2'

    # pred = '\\frac{34}{16}+\\frac{\\sqrt{1358}}{16}', gt = '4'
    # pred = '1', gt = '1\\\\sqrt{19}'

    # pred = "(0.6,2.6667]"
    # gt = "(\\frac{3}{5},\\frac{8}{3}]"

    gt = "x+2n+1"
    pred = "x+1"

    print(math_equal(pred, gt, timeout=True))


if __name__ == "__main__":
    _test_math_equal()



# Copyright 2025 the LlamaFactory team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import MethodType
from typing import TYPE_CHECKING, Any, Dict

import torch
from peft import PeftModel
from transformers import (
    PreTrainedModel,
    PreTrainedTokenizerBase,
    is_torch_npu_available,
)

from loguru import logger
from siirl.models.model_utils.visual import (
    get_image_seqlen,
    get_patch_size,
    get_vision_feature_select_strategy,
)

if TYPE_CHECKING:
    from transformers import PretrainedConfig, PreTrainedTokenizer, ProcessorMixin

    from siirl.utils.params import ModelArguments


def patch_tokenizer(tokenizer: "PreTrainedTokenizer", model_args: "ModelArguments", config: "PretrainedConfig") -> None:
    if "PreTrainedTokenizerBase" not in str(tokenizer._pad.__func__):
        tokenizer._pad = MethodType(PreTrainedTokenizerBase._pad, tokenizer)

    if model_args.model_max_length is not None and tokenizer.model_max_length != model_args.model_max_length:
        tokenizer.model_max_length = model_args.model_max_length

    if model_args.new_special_tokens is not None:
        num_added_tokens = tokenizer.add_special_tokens(
            dict(additional_special_tokens=model_args.new_special_tokens),
            replace_additional_special_tokens=False,
        )
        logger.info("Add {} to special tokens.".format(",".join(model_args.new_special_tokens)))
        if num_added_tokens > 0 and not model_args.resize_vocab:
            model_args.resize_vocab = True
            logger.warning("New tokens have been added, changed `resize_vocab` to True.")

    if "InternVL" in config.architectures[0]:

        def eos_token_id_patch(self):
            return self.super().eos_token_id

        tokenizer.__class__.eos_token_id = property(eos_token_id_patch)
        tokenizer.eos_token = "<|im_end|>"
        tokenizer.eos_token_id = tokenizer.convert_tokens_to_ids("<|im_end|>")


def patch_processor(
    processor: "ProcessorMixin",
    config: "PretrainedConfig",
    tokenizer: "PreTrainedTokenizer",
    model_args: "ModelArguments",
) -> None:
    setattr(processor, "tokenizer", tokenizer)
    if getattr(config, "vision_config", None) is not None:  # visual models
        setattr(processor, "image_seqlen", get_image_seqlen(config))
        setattr(processor, "patch_size", get_patch_size(config, processor))
        setattr(processor, "image_max_pixels", model_args.image_max_pixels)
        setattr(processor, "image_min_pixels", model_args.image_min_pixels)
        setattr(processor, "video_max_pixels", model_args.video_max_pixels)
        setattr(processor, "video_min_pixels", model_args.video_min_pixels)
        setattr(processor, "video_fps", model_args.video_fps)
        setattr(processor, "video_maxlen", model_args.video_maxlen)
        setattr(
            processor,
            "vision_feature_select_strategy",
            get_vision_feature_select_strategy(config, processor),
        )



# Copyright 2025, Shanghai Innovation Institute.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import asyncio
import ray
import torch
import numpy as np
from tensordict import TensorDict
from typing import Optional, Dict, Any

# Assume siirl is in PYTHONPATH
from siirl.workers.databuffer import DataBuffer
from siirl.workers.databuffer import DataProto


def compare_dataprotos(dp1: Optional[DataProto], dp2: Optional[DataProto], check_meta=True) -> bool:
    """A more robust DataProto comparison function for testing."""
    if dp1 is None and dp2 is None:
        return True
    if dp1 is None or dp2 is None:
        return False

    if check_meta and dp1.meta_info != dp2.meta_info:
        return False

    # Compare batch (TensorDict)
    batch1_is_none = dp1.batch is None or not dp1.batch.keys()
    batch2_is_none = dp2.batch is None or not dp2.batch.keys()

    if batch1_is_none != batch2_is_none:
        return False

    if not batch1_is_none:
        if dp1.batch.batch_size != dp2.batch.batch_size:
            return False
        if set(dp1.batch.keys()) != set(dp2.batch.keys()):
            return False
        for key in dp1.batch.keys():
            if not torch.equal(dp1.batch[key], dp2.batch[key]):
                return False

    # Compare non_tensor_batch (Dict[str, np.ndarray])
    if set(dp1.non_tensor_batch.keys()) != set(dp2.non_tensor_batch.keys()):
        return False
    for key in dp1.non_tensor_batch.keys():
        if not np.array_equal(dp1.non_tensor_batch[key], dp2.non_tensor_batch[key]):
            return False

    return True


class TestDataBuffer(unittest.IsolatedAsyncioTestCase):
    """
    Unit tests for the new logic in DataBuffer:
    - put: Executes concat under specific conditions.
    - get: Returns either an ObjectRef or a sub-shard of a DataProto.
    """

    @classmethod
    def setUpClass(cls):
        if not ray.is_initialized():
            ray.init(num_cpus=2, ignore_reinit_error=True, logging_level="error")

    @classmethod
    def tearDownClass(cls):
        if ray.is_initialized():
            ray.shutdown()

    async def asyncSetUp(self):
        """Create a new, clean DataBuffer actor for each test."""
        # Use a unique name to avoid conflicts between tests
        actor_name = f"TestBuffer_{asyncio.get_running_loop().time()}"
        # Dynamically create an Actor class from the DataBuffer class
        RemoteDataBuffer = ray.remote(DataBuffer)
        self.buffer = RemoteDataBuffer.options(name=actor_name, lifetime="detached").remote(buffer_id=0)  # type: ignore
        # Ensure the actor has started
        self.assertTrue(await self.buffer.put.remote("__init__", self._create_dp({})))

    async def asyncTearDown(self):
        """Destroy the actor after each test."""
        ray.kill(self.buffer, no_restart=True)
        # Wait briefly to ensure the actor is properly cleaned up
        await asyncio.sleep(0.1)

    def _create_dp(self, tensor_data: Dict[str, Any], non_tensor_data: Optional[Dict[str, Any]] = None, meta_info: Optional[Dict] = None) -> DataProto:
        """A simple helper function to create a DataProto for testing."""
        tensors = {k: torch.tensor(v) for k, v in tensor_data.items()} if tensor_data else {}
        non_tensors = {k: np.array(v) for k, v in non_tensor_data.items()} if non_tensor_data else {}

        batch_size = []
        if tensors:
            # Use the shape of the first tensor to determine the batch size
            batch_size = [tensors[list(tensors.keys())[0]].shape[0]]

        td = TensorDict(tensors, batch_size=batch_size) if tensors else None
        return DataProto(batch=td, non_tensor_batch=non_tensors, meta_info=meta_info or {})

    # === `put` method tests ===

    async def test_put_concat_two_dataprotos(self):
        """Core test: Verify the concat operation of put on two DataProtos."""
        key = "concat_key"
        dp1 = self._create_dp({"data": [[1, 1], [2, 2]]}, meta_info={"id": "first_put"})
        dp2 = self._create_dp({"data": [[3, 3]]}, meta_info={"id": "second_put"})

        # First put
        self.assertTrue(await self.buffer.put.remote(key, dp1))
        # Second put, should trigger concat
        self.assertTrue(await self.buffer.put.remote(key, dp2))

        # Retrieve the result for verification
        retrieved_dp = await self.buffer.get.remote(key, 0, 1)

        # Expected concat result
        expected_tensors = {"data": [[1, 1], [2, 2], [3, 3]]}
        # DataProto.concat uses the meta_info of the first element in the list.
        # In DataBuffer, it's `concat([existing_value, value])`, so the meta_info should be from dp1.
        expected_meta = {"id": "first_put"}
        expected_dp = self._create_dp(expected_tensors, meta_info=expected_meta)

        self.assertTrue(compare_dataprotos(retrieved_dp, expected_dp), "DataProtos after concat do not match expected result")

    async def test_put_replace_if_existing_is_objectref(self):
        """Test: If the original value is an ObjectRef, putting a DataProto should perform a replacement."""
        key = "replace_ref_key"
        dp_for_ref = self._create_dp({"ref_data": [10]})
        obj_ref = ray.put(dp_for_ref)

        # Store the wrapped ObjectRef ([obj_ref])
        self.assertTrue(await self.buffer.put.remote(key, [obj_ref]))

        dp_new = self._create_dp({"new_data": [20]})
        # Store a DataProto, which should replace the previous ObjectRef
        self.assertTrue(await self.buffer.put.remote(key, dp_new))

        retrieved_dp = await self.buffer.get.remote(key, 0, 1)
        self.assertIsInstance(retrieved_dp, DataProto)
        self.assertTrue(compare_dataprotos(retrieved_dp, dp_new))

    async def test_put_replace_if_new_value_is_objectref(self):
        """Test: If the new value is an ObjectRef, it should always perform a replacement."""
        key = "replace_with_ref_key"
        dp_initial = self._create_dp({"initial_data": [30]})
        self.assertTrue(await self.buffer.put.remote(key, dp_initial))  # Store a DataProto

        dp_for_ref = self._create_dp({"ref_data": [40]})
        obj_ref = ray.put(dp_for_ref)
        # Store a wrapped ObjectRef, which should replace the previous DataProto
        self.assertTrue(await self.buffer.put.remote(key, [obj_ref]))

        retrieved_ref = await self.buffer.get.remote(key, 0, 1)
        # Now get should return the ObjectRef directly
        self.assertIsInstance(retrieved_ref, ray.ObjectRef, f"Expected ObjectRef, but got {type(retrieved_ref)}")

        retrieved_dp = await retrieved_ref  # type: ignore
        self.assertTrue(compare_dataprotos(retrieved_dp, dp_for_ref))

    # === `get` method tests ===

    async def test_get_returns_objectref_if_stored(self):
        """Test: When an ObjectRef is stored, get returns it directly."""
        key = "get_ref_key"
        dp = self._create_dp({"data": [101]})
        obj_ref = ray.put(dp)
        # Store the wrapped ObjectRef
        self.assertTrue(await self.buffer.put.remote(key, [obj_ref]))

        retrieved_item = await self.buffer.get.remote(key)
        self.assertIsInstance(retrieved_item, ray.ObjectRef, f"Expected ObjectRef, but got {type(retrieved_item)}")
        self.assertEqual(retrieved_item, obj_ref)

    async def test_get_reshards_stored_dataprotos_chunk(self):
        """Test: When a DataProto (shard) is stored, get returns a re-sliced sub-shard."""
        key = "get_reshard_key"
        # Simulate a shard stored in the buffer
        dp_chunk = self._create_dp({"data": [1, 2, 3, 4, 5, 6]}, meta_info={"id": "chunk_1"})
        self.assertTrue(await self.buffer.put.remote(key, dp_chunk))

        # Target DAG worker group size is 3
        target_world_size = 3

        # Request sub-shard for rank 0
        sub_chunk_0 = await self.buffer.get.remote(key, 0, target_world_size)
        expected_0 = self._create_dp({"data": [1, 2]}, meta_info={"id": "chunk_1"})
        self.assertTrue(compare_dataprotos(sub_chunk_0, expected_0))

        # Request sub-shard for rank 1
        sub_chunk_1 = await self.buffer.get.remote(key, 1, target_world_size)
        expected_1 = self._create_dp({"data": [3, 4]}, meta_info={"id": "chunk_1"})
        self.assertTrue(compare_dataprotos(sub_chunk_1, expected_1))

        # Request sub-shard for rank 2
        sub_chunk_2 = await self.buffer.get.remote(key, 2, target_world_size)
        expected_2 = self._create_dp({"data": [5, 6]}, meta_info={"id": "chunk_1"})
        self.assertTrue(compare_dataprotos(sub_chunk_2, expected_2))

    async def test_get_returns_none_for_nonexistent_key(self):
        """Test: When the key does not exist, get returns None."""
        retrieved_item = await self.buffer.get.remote("non_existent_key")
        self.assertIsNone(retrieved_item)

    # === `pop` method tests ===

    async def test_pop_removes_and_returns_dataprotos(self):
        """Test: pop can remove and return a DataProto."""
        key = "pop_dp_key"
        dp = self._create_dp({"data": [1, 2, 3]})

        # Store data
        await self.buffer.put.remote(key, dp)

        # Confirm data exists
        self.assertIsNotNone(await self.buffer.get.remote(key))

        # pop operation
        popped_item = await self.buffer.pop.remote(key)
        self.assertIsInstance(popped_item, DataProto)
        self.assertTrue(compare_dataprotos(popped_item, dp))

        # Confirm data has been removed
        self.assertIsNone(await self.buffer.get.remote(key))

    async def test_pop_removes_and_returns_objectref(self):
        """Test: pop can remove and return an ObjectRef."""
        key = "pop_ref_key"
        dp_for_ref = self._create_dp({"ref_data": [100]})
        obj_ref = ray.put(dp_for_ref)

        # Store the wrapped ObjectRef
        await self.buffer.put.remote(key, [obj_ref])

        # Confirm data exists
        self.assertIsNotNone(await self.buffer.get.remote(key))

        # pop operation
        popped_item = await self.buffer.pop.remote(key)
        self.assertIsInstance(popped_item, ray.ObjectRef)
        self.assertEqual(popped_item, obj_ref)

        # Confirm data has been removed
        self.assertIsNone(await self.buffer.get.remote(key))

    async def test_pop_returns_none_for_nonexistent_key(self):
        """Test: When the key does not exist, pop returns None."""
        popped_item = await self.buffer.pop.remote("non_existent_key_for_pop")
        self.assertIsNone(popped_item)


if __name__ == "__main__":
    unittest.main()



# Copyright 2025, Shanghai Innovation Institute. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List
from siirl.workers.dag.node import NodeRole


class DAGInitializationError(Exception):
    """Custom exception for failures during DAGWorker initialization."""

    pass


class DAGConstants:
    """Centralized constants to improve maintainability and avoid magic strings."""

    # Worker role mapping
    WORKER_ROLE_MAPPING: Dict[NodeRole, str] = {
        NodeRole.ACTOR: "actor",
        NodeRole.ROLLOUT: "rollout",
        NodeRole.REFERENCE: "ref",
    }
    # Configuration keys
    INTERN_CONFIG: str = "intern_config"
    # Framework strategy names
    FSDP_STRATEGIES: List[str] = ["fsdp", "fsdp2"]
    MEGATRON_STRATEGY: str = "megatron"
    # Metric group order
    METRIC_GROUP_ORDER = ["step", "training", "actor", "critic", "perf", "response_length", "response", "prompt_length", "prompt", "global_seqlen", "timing_s", "timing_per_token_ms", "perf/total_num_tokens", "perf/time_per_step", "perf/throughput"]



# Copyright (c) 2025, Shanghai Innovation Institute. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import os
from typing import Dict, Type
import torch
import ray
import torch

import torch.distributed as dist
from loguru import logger

from siirl.dataloader import DataLoaderNode
from siirl.models.loader import load_tokenizer
from siirl.workers.base_worker import Worker
from siirl.scheduler.reward import create_reward_manager
from siirl.workers.dag.node import NodeRole, NodeType
from siirl.workers.dag_worker.constants import DAGConstants
from siirl.utils.extras.device import get_device_name, get_nccl_backend
device_name = get_device_name()

class InitializationMixin:
    """Handles the initialization and setup logic for the DAGWorker."""

    from typing import Dict, List, Optional, Type, Any
    from torch.distributed import ProcessGroup
    from siirl.utils.params import SiiRLArguments
    from siirl.scheduler.process_group_manager import ProcessGroupManager
    from siirl.models.loader import TokenizerModule
    from siirl.workers.dag import TaskGraph
    from siirl.workers.dag.node import Node, NodeRole
    from siirl.workers.base_worker import Worker
    from siirl.utils.logger.tracking import Tracking

    # Attributes from DAGWorker's __init__
    config: SiiRLArguments
    process_group_manager: ProcessGroupManager
    taskgraph_mapping: Dict[int, TaskGraph]
    data_buffers: List["ray.actor.ActorHandle"]
    enable_perf: bool
    workers: Dict[str, Worker]
    agent_group_worker: Dict[int, Dict[NodeRole, Worker]]
    agent_group_process_group: Dict[int, Dict[NodeRole, ProcessGroup]]
    process_groups: Dict[str, ProcessGroup]
    tokenizer_mapping: Dict[str, TokenizerModule]
    logger: Optional[Tracking]

    # Attributes initialized within this mixin
    _rank: int
    taskgraph: TaskGraph
    _gather_group: Optional[ProcessGroup]
    first_rollout_node: Node
    dataloader: "DataLoaderNode"
    val_reward_fn: Any
    reward_fn: Any
    kl_ctrl_in_reward: Optional[Any]
    validate_tokenizer: Any
    role_worker_mapping: Dict[NodeRole, Type[Worker]]

    def _initialize_worker(self):
        """Orchestrates the ordered initialization of all worker components."""
        self._rank = self._get_and_validate_rank()
        self.taskgraph = self._get_taskgraph_for_rank(self.taskgraph_mapping)
        self._setup_distributed_environment()
        self._initialize_core_components()
        self._initialize_node_workers()

        if self._rank == 0:
            logger.info("Rank 0: Initializing tracking logger...")
            from siirl.utils.logger.tracking import Tracking

            self.logger = Tracking(
                project_name=self.config.trainer.project_name,
                experiment_name=self.config.trainer.experiment_name,
                default_backend=self.config.trainer.logger,
                config=self.config.to_dict(),
            )
            if self.enable_perf:
                logger.warning("Performance tracking is enabled. This may impact training speed.")

    def _get_and_validate_rank(self) -> int:
        """Retrieves and validates the worker's rank from the environment."""
        rank_str = os.environ.get("RANK")
        if rank_str is None:
            raise ValueError("Environment variable 'RANK' is not set. This is required for distributed setup.")
        try:
            return int(rank_str)
        except ValueError:
            raise ValueError(f"Invalid RANK format: '{rank_str}'. Must be an integer.")

    def _get_taskgraph_for_rank(self, taskgraph_mapping: Dict[int, "TaskGraph"]) -> "TaskGraph":
        """Retrieves the TaskGraph for the current rank from the provided mapping."""
        if self._rank not in taskgraph_mapping:
            raise ValueError(f"Rank {self._rank} not found in the provided taskgraph_mapping.")
        taskgraph = taskgraph_mapping[self._rank]
        from siirl.workers.dag import TaskGraph

        if not isinstance(taskgraph, TaskGraph):
            raise TypeError(f"Object for rank {self._rank} must be a TaskGraph, but got {type(taskgraph).__name__}.")
        logger.info(f"Rank {self._rank} assigned to TaskGraph with ID {taskgraph.graph_id}.")
        return taskgraph

    def _setup_distributed_environment(self):
        """Initializes the default process group and all required subgroups."""
        # gloo_socket_ifname = 'bond0'
        # os.environ["GLOO_SOCKET_IFNAME"] = gloo_socket_ifname
        # os.environ["GLOO_LOG_LEVEL"] = "DEBUG"
        import torch.distributed as dist

        if not dist.is_initialized():
            backend = f"{get_nccl_backend()}" if self.world_size >= self.config.dag.backend_threshold else f"cpu:gloo,{get_device_name()}:{get_nccl_backend()}"
            logger.info(f"Rank {self._rank}: Initializing world size {self.world_size} default process group with '{backend}' backend.")
            dist.init_process_group(backend=backend)

        if device_name == "npu":
            # For NPU, metrics aggregation requires the hccl backend for device-to-device communication.
            # This group is created regardless of world size for NPU environments.
            gather_backend = get_nccl_backend()
            self._gather_group = dist.new_group(backend=gather_backend)
        else:
            # For GPU, the original logic is preserved for backward compatibility.
            # The gather group is only created if world_size < 256.
            if self.world_size < 256:
                self._gather_group = dist.new_group(backend="gloo")
            else:
                self._gather_group = None
        self._build_all_process_groups()
        self._resolve_taskgraph_process_groups()
        # Ensure all ranks have finished group creation before proceeding.
        dist.barrier(self._gather_group)
        logger.info(f"Rank {self._rank}: Distributed environment setup complete.")

    def _build_all_process_groups(self):
        """Builds all process groups defined in the ProcessGroupManager."""
        import torch.distributed as dist

        group_specs = self.process_group_manager.get_all_specs()
        if not group_specs:
            logger.warning("No process group specifications found in ProcessGroupManager.")
            return

        for name, spec in group_specs.items():
            if not isinstance(spec, dict) or not (ranks := spec.get("ranks")):
                logger.warning(f"Skipping group '{name}' due to invalid spec or missing 'ranks'.")
                continue
            self.process_groups[name] = dist.new_group(ranks=ranks)
        logger.debug(f"Rank {self._rank}: Created {len(self.process_groups)} custom process groups.")

    def _resolve_taskgraph_process_groups(self):
        """Identifies and caches process groups relevant to this worker's TaskGraph."""
        self.inference_group_name_set = self.process_group_manager.get_process_group_for_node_type_in_subgraph(self.taskgraph.graph_id, NodeType.MODEL_INFERENCE.value)
        self.train_group_name_set = self.process_group_manager.get_process_group_for_node_type_in_subgraph(self.taskgraph.graph_id, NodeType.MODEL_TRAIN.value)

    def _initialize_core_components(self):
        """Initializes shared components like tokenizers, data loaders, and reward functions."""
        self._setup_tokenizers()
        self._setup_dataloader_and_reward()
        self._setup_role_worker_mapping()

    def _setup_tokenizers(self):
        """Initializes and caches tokenizers for all models in the task graph."""
        model_nodes = [node for node in self.taskgraph.nodes.values() if node.node_type in [NodeType.MODEL_TRAIN, NodeType.MODEL_INFERENCE]]
        if not model_nodes:
            logger.warning("No model nodes found in the task graph. Tokenizer setup will be skipped.")
            return

        for node in model_nodes:
            agent_key = self._generate_agent_group_key(node)
            if agent_key not in self.tokenizer_mapping:
                # Add robust check for missing configuration.
                intern_config = node.config.get(DAGConstants.INTERN_CONFIG)
                if not intern_config or not (model_dict := getattr(intern_config, "model", None)):
                    logger.warning(f"Node {node.node_id} is missing model config. Skipping tokenizer setup for it.")
                    continue

                tokenizer_module = load_tokenizer(model_args=model_dict)
                if tokenizer := tokenizer_module.get("tokenizer"):
                    tokenizer.padding_side = "left"  # Required for most causal LM generation
                self.tokenizer_mapping[agent_key] = tokenizer_module
        logger.info(f"Rank {self._rank}: Initialized {len(self.tokenizer_mapping)} tokenizer(s).")

    def _setup_dataloader_and_reward(self):
        """Initializes the data loader and reward functions."""
        rollout_nodes = [n for n in self.taskgraph.nodes.values() if n.node_type == NodeType.MODEL_INFERENCE]
        if not rollout_nodes:
            raise ValueError("At least one MODEL_INFERENCE node is required for dataloader and reward setup.")
        self.first_rollout_node = rollout_nodes[0]

        pg_assignment = self.process_group_manager.get_node_assignment(self.first_rollout_node.node_id)
        if not (process_group_name := pg_assignment.get("process_group_name")):
            raise ValueError(f"Process group name not found for the first rollout node {self.first_rollout_node.node_id}.")

        self.dataloader_process_group = self.process_groups.get(process_group_name)
        if self.dataloader_process_group is None:
            raise ValueError(f"Could not find process group '{process_group_name}' in the created groups.")

        self.dataloader_tensor_model_parallel_size = self.first_rollout_node.config[DAGConstants.INTERN_CONFIG].rollout.tensor_model_parallel_size

        self.dataloader = DataLoaderNode(
            node_id="dataloader",
            global_config=self.config,
            config={
                "group_world_size": dist.get_world_size(self.dataloader_process_group),
                "group_rank": dist.get_rank(self.dataloader_process_group),
                "group_parallel_size": self.dataloader_tensor_model_parallel_size,
                "num_loader_workers": self.config.data.num_loader_workers,
                "auto_repeat": self.config.data.auto_repeat,
            },
        )

        self.validate_tokenizer = next(iter(self.tokenizer_mapping.values()), {}).get("tokenizer")
        if not self.validate_tokenizer:
            logger.warning("No tokenizer loaded; reward functions might fail or use a default one.")

        self.val_reward_fn = create_reward_manager(self.config, self.validate_tokenizer, num_examine=1)
        self.reward_fn = create_reward_manager(self.config, self.validate_tokenizer, num_examine=0, **self.config.reward_model.reward_kwargs)

        if self.config.algorithm.use_kl_in_reward:
            from siirl.workers.dag_worker import core_algos

            self.kl_ctrl_in_reward = core_algos.get_kl_controller(self.config.algorithm.kl_ctrl)

        # TODO: support multi-agent environment

    def _get_worker_classes(self, strategy: str) -> Dict[NodeRole, Type[Worker]]:
        """Dynamically imports worker classes based on the specified strategy."""
        if strategy in DAGConstants.FSDP_STRATEGIES:
            from siirl.workers.fsdp_workers import ActorRolloutRefWorker, AsyncActorRolloutRefWorker, CriticWorker, RewardModelWorker

            actor_cls = AsyncActorRolloutRefWorker if self.config.actor_rollout_ref.rollout.mode == "async" else ActorRolloutRefWorker
            return {NodeRole.ACTOR: actor_cls, NodeRole.ROLLOUT: actor_cls, NodeRole.REFERENCE: actor_cls, NodeRole.CRITIC: CriticWorker, NodeRole.REWARD: RewardModelWorker}
        elif strategy == DAGConstants.MEGATRON_STRATEGY:
            from siirl.workers.megatron_workers import ActorRolloutRefWorker, CriticWorker, RewardModelWorker

            return {NodeRole.ACTOR: ActorRolloutRefWorker, NodeRole.ROLLOUT: ActorRolloutRefWorker, NodeRole.REFERENCE: ActorRolloutRefWorker, NodeRole.CRITIC: CriticWorker, NodeRole.REWARD: RewardModelWorker}
        raise NotImplementedError(f"Strategy '{strategy}' is not supported.")

    def _setup_role_worker_mapping(self):
        """Creates a mapping from NodeRole to the corresponding Worker implementation class."""
        self.role_worker_mapping: Dict[NodeRole, Type[Worker]] = {}
        # Actor/Ref/Rollout/Critic workers
        actor_strategy = self.config.actor_rollout_ref.actor.strategy
        self.role_worker_mapping.update(self._get_worker_classes(actor_strategy))

        # Reward model worker (if enabled)
        if self.config.reward_model.enable:
            reward_strategy = self.config.reward_model.strategy
            reward_workers = self._get_worker_classes(reward_strategy)
            if NodeRole.REWARD in reward_workers:
                self.role_worker_mapping[NodeRole.REWARD] = reward_workers[NodeRole.REWARD]
            else:
                logger.warning(f"Reward model is enabled, but no worker found for role REWARD with strategy {reward_strategy}.")

        self._log_role_worker_mapping()

    def _log_role_worker_mapping(self):
        """Logs the final role-to-worker mapping for setup verification."""
        if not self.role_worker_mapping:
            logger.error("Role-to-worker mapping is empty after setup. This will cause execution failure.")
            return

        logger.debug("--- [Role -> Worker Class] Mapping ---")
        max_len = max((len(r.name) for r in self.role_worker_mapping.keys()), default=0)
        for role, worker_cls in sorted(self.role_worker_mapping.items(), key=lambda item: item[0].name):
            logger.debug(f"  {role.name:<{max_len}} => {worker_cls.__name__} (from {inspect.getmodule(worker_cls).__name__})")
        logger.debug("--------------------------------------")

    def _initialize_node_workers(self):
        """Instantiates worker objects for all nodes in the task graph."""
        for node in self.taskgraph.nodes.values():
            if not self._should_create_worker(node):
                continue

            worker_cls = self.role_worker_mapping.get(node.node_role)
            if not worker_cls:
                logger.warning(f"No worker class found for role {node.node_role.name}. Skipping node {node.node_id}.")
                continue

            node_worker_key = self._generate_node_worker_key(node)
            if node_worker_key in self.workers:
                continue

            try:
                node_process_group = self._get_node_process_group(node)
                config = node.config.get(DAGConstants.INTERN_CONFIG)
                if hasattr(config, "actor") and hasattr(config.actor, "optim"):
                    config.actor.optim.total_training_steps = self.dataloader.total_training_steps
                elif hasattr(config, "optim"):
                    config.optim.total_training_steps = self.dataloader.total_training_steps
                worker_args = {"config": config, "process_group": node_process_group}
                if node.node_role in DAGConstants.WORKER_ROLE_MAPPING:
                    worker_args["role"] = DAGConstants.WORKER_ROLE_MAPPING[node.node_role]

                worker_instance = worker_cls(**worker_args)
                self.workers[node_worker_key] = worker_instance
                self.agent_group_worker[node.agent_group][node.node_role] = worker_instance
                self.agent_group_process_group[node.agent_group][node.node_role] = node_process_group
                logger.success(f"Rank {self._rank}: Successfully created worker '{worker_cls.__name__}' for node: {node.node_id}")
                
                # note all agents share same critic in multi-agent(Marft)
                if node.node_role == NodeRole.CRITIC and node.agent_group != 0:
                    for agent in range(node.agent_group):
                        self.agent_group_worker[agent][node.node_role] = worker_instance

            except Exception as e:
                #  Explicitly log the failing node and worker class, then re-raise
                # the exception to prevent silent failures.
                logger.error(f"Failed to create worker for node {node.node_id} with class {worker_cls.__name__}.", exc_info=True)
                raise RuntimeError(f"Worker instantiation failed for node {node.node_id}") from e

    def _generate_node_worker_key(self, node: Node) -> str:
        """Generates a unique string key for a node's worker instance."""
        return f"{node.agent_group}_{node.node_type.value}_{node.node_role.value}"

    def _generate_agent_group_key(self, node: Node) -> str:
        """Generates a unique key for an agent group, used for caching (e.g., tokenizers)."""
        return f"group_key_{node.agent_group}"

    def _should_create_worker(self, node: Node) -> bool:
        """Determines if a worker instance should be created for a given graph node."""
        return node.node_type in [NodeType.MODEL_TRAIN, NodeType.MODEL_INFERENCE] and node.node_role in self.role_worker_mapping

    def _get_node_process_group(self, node: Node) -> ProcessGroup:
        """Retrieves the PyTorch ProcessGroup assigned to a specific graph node."""
        assignment = self.process_group_manager.get_node_assignment(node.node_id)
        if not (assignment and (name := assignment.get("process_group_name"))):
            raise ValueError(f"Process group assignment or name not found for node {node.node_id}.")

        pg = self.process_groups.get(name)
        if pg is None:
            raise ValueError(f"Process group '{name}' for node {node.node_id} was not created or found.")
        return pg

    def _get_node(self, role: NodeRole, agent_group: int) -> Node:
        """
        Finds and returns a specific node from the task graph based on its role
        and agent group.
        """
        found_node = next((node for node in self.taskgraph.nodes.values() if node.node_role == role and node.agent_group == agent_group), None)

        if found_node is None:
            raise RuntimeError(f"Could not find a node with role {role.name} for agent_group {agent_group}")
        return found_node

    def _get_node_dp_info(self, node: Node) -> tuple[int, int, int, int]:
        """Calculates Data Parallel (DP) and Tensor Parallel (TP) info for a node."""
        reference_node = node
        if node.node_type == NodeType.COMPUTE:
            # If the node is a COMPUTE type, find its true data source ancestor.
            ancestor = self._find_first_non_compute_ancestor(node.node_id)
            if ancestor:
                reference_node = ancestor
            else:
                # If no non-COMPUTE ancestor is found, it's a critical error.
                raise RuntimeError(f"Could not find any non-COMPUTE ancestor for COMPUTE node '{node.node_id}'. Please check your DAG graph configuration.")

        if reference_node.node_type == NodeType.COMPUTE:
            group_world_size = self.config.trainer.n_gpus_per_node * self.config.trainer.nnodes
            group_rank = dist.get_rank()
        else:
            process_group = self._get_node_process_group(reference_node)
            group_world_size = dist.get_world_size(process_group)
            group_rank = dist.get_rank(process_group)

        tp_size = 1
        if intern_config := reference_node.config.get(DAGConstants.INTERN_CONFIG):
            if reference_node.node_type == NodeType.MODEL_INFERENCE:
                tp_size = intern_config.rollout.tensor_model_parallel_size
            # TODO: Add support for Megatron strategy, reading from its specific model config.

        if group_world_size % tp_size != 0:
            raise ValueError(f"Configuration error for node {node.node_id}: Group world size ({group_world_size}) is not divisible by tensor parallel size ({tp_size}). Check your parallel configuration.")
        dp_size = group_world_size // tp_size
        dp_rank = group_rank // tp_size
        tp_rank = group_rank % tp_size
        return dp_size, dp_rank, tp_rank, tp_size

    def log_ray_actor_info(self):
        """Logs detailed information about the Ray actor's context for debugging."""
        try:
            ctx = ray.get_runtime_context()
            logger.debug(f"Ray Actor Context for Rank {self._rank}: ActorID={ctx.get_actor_id()}, JobID={ctx.get_job_id()}, NodeID={ctx.get_node_id()}, PID={os.getpid()}")
        except RuntimeError:
            logger.warning(f"Rank {self._rank}: Not running in a Ray actor context.")

    def init_model(self):
        """Initializes models for all workers and sets up sharding managers where applicable."""
        logger.info("Initializing models for all worker nodes...")
        have_init_workers = set()
        for node in self.taskgraph.nodes.values():
            if self._should_create_worker(node):
                node_worker = self.workers[self._generate_node_worker_key(node)]
                if not isinstance(node_worker, Worker):
                    raise TypeError(f"Invalid worker type for node {node.node_id}: {type(node_worker).__name__}")
                if self._generate_node_worker_key(node) in have_init_workers:
                    logger.warning(f"Rank {self._rank}: Worker {self._generate_node_worker_key(node)} for node {node.node_id} already initialized. Skipping.")
                    continue
                node_worker.init_model()
                have_init_workers.add(self._generate_node_worker_key(node))
                if node.node_role == NodeRole.ROLLOUT and node.config['intern_config'].rollout.mode == 'async':
                    self.rollout_mode = 'async'
                    self.zmq_address = node_worker.get_zeromq_address()
        logger.success("All worker models initialized.")

        logger.info(f"Setting up sharding managers {self.config.actor_rollout_ref.rollout.name} ...")
        for agent_group, worker_dict in self.agent_group_worker.items():
            if NodeRole.ACTOR in worker_dict and NodeRole.ROLLOUT in worker_dict:
                try:
                    self._setup_sharding_manager(agent_group, worker_dict)
                except Exception as e:
                    logger.error(f"Failed to set up sharding manager for agent group {agent_group}: {e}", exc_info=True)
                    raise
        logger.info("All models and sharding managers initialized successfully.")

    def _setup_sharding_manager(self, agent_group: int, worker_dict: Dict[NodeRole, Worker]):
        """Configures the sharding manager to sync weights between FSDP and vLLM."""
        actor_worker = worker_dict[NodeRole.ACTOR]
        rollout_worker = worker_dict[NodeRole.ROLLOUT]
        rollout_pg = self.agent_group_process_group[agent_group][NodeRole.ROLLOUT]

        parallel_config = {"rollout_parallel_size": rollout_worker.config.rollout.tensor_model_parallel_size, "rollout_world_size": dist.get_world_size(rollout_pg), "rollout_rank": dist.get_rank(rollout_pg)}

        if self.config.actor_rollout_ref.rollout.name == "vllm":
            from siirl.workers.sharding_manager.fsdp_vllm import MultiAgentFSDPVLLMShardingManager

            sharding_manager_cls = MultiAgentFSDPVLLMShardingManager
            sharding_manager = sharding_manager_cls(
                module=actor_worker.actor_module_fsdp,
                inference_engine=rollout_worker.rollout.inference_engine,
                model_config=actor_worker.actor_model_config,
                parallel_config=parallel_config,
                full_params="hf" in rollout_worker.config.rollout.load_format,
                offload_param=getattr(actor_worker, "_is_offload_param", False)
            )
        elif self.config.actor_rollout_ref.rollout.name == "sglang":
            from siirl.workers.sharding_manager.fsdp_sglang import MultiAgentFSDPSGLangShardingManager

            sharding_manager_cls = MultiAgentFSDPSGLangShardingManager
            tp_size = parallel_config.get("rollout_parallel_size")
            world_size = parallel_config.get("rollout_world_size")
            rollout_device_mesh = torch.distributed.init_device_mesh(
                device_name, mesh_shape=(world_size // tp_size, tp_size), mesh_dim_names=["dp", "infer_tp"]
            )
            sharding_manager = sharding_manager_cls(
                module=actor_worker.actor_module_fsdp,
                inference_engine=rollout_worker.rollout.inference_engine,
                model_config=actor_worker.actor_model_config,
                device_mesh=rollout_device_mesh,
                rollout_config=rollout_worker.config.rollout,
                full_params="hf" in rollout_worker.config.rollout.load_format,
                offload_param=getattr(actor_worker, "_is_offload_param", False),
                multi_stage_wake_up=rollout_worker.config.rollout.multi_stage_wake_up
            )
        else:
            raise NotImplementedError(f"{self.config.actor_rollout_ref.rollout.name} not supported")
        rollout_worker.set_rollout_sharding_manager(sharding_manager)
        logger.debug(f"Set up {sharding_manager_cls.__name__}  for agent group {agent_group}.")

    def init_graph(self):
        # this is needed by async rollout manager
        self._set_node_executables()
        self.init_model()
        self._load_checkpoint()
        # Ensure all models are initialized and checkpoints are loaded before starting.
        dist.barrier(self._gather_group)    
        
        
    def set_async_rollout_manager(self, async_rollout_manager):
        self._async_rollout_manager = async_rollout_manager
        
    def get_zeromq_address(self):
        return self.zmq_address




# Copyright 2025, Shanghai Innovation Institute.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import asyncio
import ray
import torch
import torch.distributed as dist
from tensordict import TensorDict
from typing import List, Optional, Dict

from siirl.workers.databuffer.data_buffer import DataBuffer
from siirl.workers.databuffer import DataProto
from tests.data_buffer.test_data_buffer import compare_dataprotos


# Mock Tokenizer for testing purposes
class MockTokenizer:
    def __init__(self, pad_token_id=0):
        self.pad_token_id = pad_token_id
        self.padding_side = "right"


@ray.remote
class MockDAGWorker:
    """
    A mock DAGWorker that only contains the logic needed to test put/get.
    It does not load a model or execute an actual computation graph.
    """

    def __init__(self, rank: int, world_size: int, data_buffers: List[ray.actor.ActorHandle]):
        self._rank = rank
        self._world_size = world_size
        self.data_buffers = data_buffers
        self.tokenizer = MockTokenizer()

        # Mock the distributed setup for the actor
        dist.init_process_group(
            backend="gloo",  # Use gloo for CPU-based testing
            init_method=f"tcp://127.0.0.1:29501",
            world_size=self._world_size,
            rank=self._rank,
        )

    async def put_data_to_buffers(self, key: str, data: DataProto, source_dp_size: int, dest_dp_size: int):
        # This is a copy of the logic from the actual DAGWorker
        data.meta_info["padding_values"] = {
            "input_ids": self.tokenizer.pad_token_id,
            "responses": self.tokenizer.pad_token_id,
            "labels": -100,
            "attention_mask": 0,
            "response_mask": 0,
        }
        data.meta_info["padding_side"] = self.tokenizer.padding_side

        if source_dp_size == dest_dp_size:
            obj_ref = ray.put(data)
            await self.data_buffers[0].put.remote(key, [obj_ref])
        else:
            num_physical_buffers = len(self.data_buffers)
            chunks = data.chunk(chunks=num_physical_buffers)
            put_futures = [buffer.put.remote(key, chunk) for buffer, chunk in zip(self.data_buffers, chunks)]
            await asyncio.gather(*put_futures)

    async def get_data_from_buffers(self, key: str, my_current_dp_rank: int, my_current_dp_world_size: int) -> Optional[DataProto]:
        # This is a copy of the logic from the actual DAGWorker
        if not self.data_buffers:
            return None

        first_item = await self.data_buffers[0].get.remote(key, my_current_dp_rank, my_current_dp_world_size)

        if first_item is None:
            return None

        if isinstance(first_item, ray.ObjectRef):
            return await first_item

        elif isinstance(first_item, DataProto):
            my_sub_chunks = [first_item]
            get_futures = [buffer.get.remote(key, my_current_dp_rank, my_current_dp_world_size) for buffer in self.data_buffers[1:]]
            other_sub_chunks = await asyncio.gather(*get_futures)

            for sub_chunk in other_sub_chunks:
                if not isinstance(sub_chunk, DataProto):
                    return None
                my_sub_chunks.append(sub_chunk)

            return DataProto.concat(my_sub_chunks)

        return None

    def get_rank(self):
        return self._rank

    def barrier(self):
        dist.barrier()


def _create_test_dp(batch_size: int, seq_len: int, meta: Optional[Dict] = None) -> DataProto:
    """Creates a sample DataProto for testing."""
    tensors = {
        "input_ids": torch.randint(1, 100, (batch_size, seq_len)),
        "attention_mask": torch.ones(batch_size, seq_len, dtype=torch.long),
    }
    td = TensorDict(tensors, batch_size=[batch_size])
    return DataProto(batch=td, meta_info=meta or {})


class TestDAGWorkerDataFlow(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        if not ray.is_initialized():
            ray.init(num_cpus=8, ignore_reinit_error=True, logging_level="error")

    @classmethod
    def tearDownClass(cls):
        if ray.is_initialized():
            ray.shutdown()

    async def test_put_get_flow_sharded(self):
        """
        Tests the data flow for sharded storage (source_dp != dest_dp).
        - A single worker (rank 0) puts a sharded DataProto.
        - All workers get their corresponding data and reconstruct it.
        """
        num_buffers = 2
        num_workers = 4
        source_dp, dest_dp = 2, 4
        key = "sharded_key"

        # 1. Setup Actors
        buffers = [DataBuffer.remote(buffer_id=i) for i in range(num_buffers)]
        workers = [MockDAGWorker.remote(rank=i, world_size=num_workers, data_buffers=buffers) for i in range(num_workers)]
        await asyncio.sleep(1)  # Wait for actors to initialize dist group

        # 2. Prepare Data
        # Total batch size should be divisible by dest_dp and num_buffers
        total_batch_size = dest_dp * num_buffers
        full_dp = _create_test_dp(total_batch_size, seq_len=10)

        # 3. Leader (worker 0) puts data
        # Data is sharded across the 2 buffers
        await workers[0].put_data_to_buffers.remote(key, full_dp, source_dp_size=source_dp, dest_dp_size=dest_dp)

        # 4. All workers synchronize
        await asyncio.gather(*[w.barrier.remote() for w in workers])

        # 5. All workers get their data slice
        get_futures = [w.get_data_from_buffers.remote(key, my_current_dp_rank=await w.get_rank.remote(), my_current_dp_world_size=dest_dp) for w in workers]
        results = await asyncio.gather(*get_futures)

        # 6. Verify results
        # In sharded mode, data is distributed in an interleaved manner. We must manually reconstruct
        # the expected shard for each worker.
        buffer_chunks = full_dp.chunk(chunks=num_buffers)
        expected_shards_by_worker = []
        for worker_rank in range(dest_dp):
            sub_chunks_for_worker = []
            for buffer_chunk in buffer_chunks:
                # The chunk from each buffer will be re-sharded for the target worker.
                sub_sub_chunks = buffer_chunk.chunk(chunks=dest_dp)
                sub_chunks_for_worker.append(sub_sub_chunks[worker_rank])
            # The worker's final data is the concatenation of the sub-shards it receives from all buffers.
            expected_shards_by_worker.append(DataProto.concat(sub_chunks_for_worker))

        self.assertEqual(len(results), len(expected_shards_by_worker))

        for i, result_dp in enumerate(results):
            self.assertIsNotNone(result_dp, f"Worker {i} received None")
            expected_dp = expected_shards_by_worker[i]
            self.assertTrue(compare_dataprotos(expected_dp, result_dp, check_meta=False), f"DataProto for worker {i} does not match the expected interleaved shard. Expected size {len(expected_dp)}, got {len(result_dp)}")

        # Cleanup
        for actor in workers + buffers:
            ray.kill(actor, no_restart=True)

    async def test_put_get_flow_object_ref(self):
        """
        Tests the data flow for ObjectRef storage (source_dp == dest_dp).
        - A single worker (rank 0) puts a DataProto, which becomes an ObjectRef.
        - All workers get and resolve the same ObjectRef.
        """
        num_buffers = 1  # Only one buffer is used in this case
        num_workers = 4
        source_dp, dest_dp = 4, 4
        key = "object_ref_key"

        # 1. Setup Actors
        buffers = [DataBuffer.remote(buffer_id=0)]
        workers = [MockDAGWorker.remote(rank=i, world_size=num_workers, data_buffers=buffers) for i in range(num_workers)]
        await asyncio.sleep(1)

        # 2. Prepare Data
        full_dp = _create_test_dp(batch_size=8, seq_len=12)

        # 3. Leader (worker 0) puts data
        await workers[0].put_data_to_buffers.remote(key, full_dp, source_dp_size=source_dp, dest_dp_size=dest_dp)

        # 4. All workers synchronize
        await asyncio.gather(*[w.barrier.remote() for w in workers])

        # 5. All workers get data and verify
        get_futures = [w.get_data_from_buffers.remote(key, my_current_dp_rank=await w.get_rank.remote(), my_current_dp_world_size=dest_dp) for w in workers]
        results = await asyncio.gather(*get_futures)

        # In the ObjectRef case, every worker gets the same data.
        for i, result_dp in enumerate(results):
            self.assertIsNotNone(result_dp, f"Worker {i} received None")
            self.assertTrue(compare_dataprotos(full_dp, result_dp, check_meta=False), f"DataProto for worker {i} does not match the original")

        # Cleanup
        for actor in workers + buffers:
            ray.kill(actor)


if __name__ == "__main__":
    unittest.main()



# Copyright 2025, Shanghai Innovation Institute.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import copy
from typing import List, Set, Dict, Tuple

# Assuming node.py, task_graph.py, task_loader.py are accessible
from siirl.workers.dag import Node, NodeType, TaskGraph
from siirl.workers.dag.task_loader import generate_structural_signature, get_all_downstream_nodes_recursive, get_all_ancestors, find_all_paths, split_single_structure, split_by_fan_out_to_exits, split_by_reconverging_paths, discover_and_split_parallel_paths


# Helper to get a set of structural signatures for a list of graphs
def get_signatures(graphs: List[TaskGraph]) -> List[str]:
    """Generates a sorted list of structural signatures for a list of graphs."""
    return sorted([generate_structural_signature(g) for g in graphs])


class TestTaskLoaderInternals(unittest.TestCase):
    """Tests for internal helper functions in task_loader.py."""

    def setUp(self):
        # Simple graph: A -> B -> C, A -> D
        #      A
        #     / \
        #    B   D
        #    |
        #    C
        self.node_a = Node("A", NodeType.DATA_LOAD)
        self.node_b = Node("B", NodeType.COMPUTE, dependencies=["A"])
        self.node_c = Node("C", NodeType.COMPUTE, dependencies=["B"])
        self.node_d = Node("D", NodeType.COMPUTE, dependencies=["A"])
        self.graph1 = TaskGraph("g1")
        self.graph1.add_nodes([self.node_a, self.node_b, self.node_c, self.node_d])
        self.graph1.build_adjacency_lists()  # Crucial for many helpers

        # Linear graph: L1 -> L2 -> L3
        self.ln1 = Node("L1", NodeType.DATA_LOAD)
        self.ln2 = Node("L2", NodeType.COMPUTE, dependencies=["L1"])
        self.ln3 = Node("L3", NodeType.MODEL_TRAIN, dependencies=["L2"])
        self.linear_graph = TaskGraph("linear")
        self.linear_graph.add_nodes([self.ln1, self.ln2, self.ln3])
        self.linear_graph.build_adjacency_lists()

        # Empty graph
        self.empty_graph = TaskGraph("empty")

    def test_generate_structural_signature(self):
        sig1 = generate_structural_signature(self.graph1)

        graph1_reordered_nodes = TaskGraph("g1_reordered")
        # Add nodes in different order but same structure
        graph1_reordered_nodes.add_node(copy.deepcopy(self.node_d))
        graph1_reordered_nodes.add_node(copy.deepcopy(self.node_c))
        graph1_reordered_nodes.add_node(copy.deepcopy(self.node_b))
        graph1_reordered_nodes.add_node(copy.deepcopy(self.node_a))
        sig2 = generate_structural_signature(graph1_reordered_nodes)
        self.assertEqual(sig1, sig2, "Signatures should match for structurally identical graphs regardless of node add order.")

        sig_empty = generate_structural_signature(self.empty_graph)
        self.assertTrue("empty_structure" in sig_empty)

        sig_linear = generate_structural_signature(self.linear_graph)
        self.assertTrue("L1" in sig_linear and "L2" in sig_linear and "L3" in sig_linear)
        self.assertTrue("e(L1->L2)" in sig_linear)

    def test_get_all_downstream_nodes_recursive(self):
        # For self.graph1 (A -> B -> C, A -> D)
        # Downstream of A should be A, B, C, D
        downstream_a = get_all_downstream_nodes_recursive(self.graph1, "A")
        self.assertSetEqual(downstream_a, {"A", "B", "C", "D"})

        # Downstream of B should be B, C
        downstream_b = get_all_downstream_nodes_recursive(self.graph1, "B")
        self.assertSetEqual(downstream_b, {"B", "C"})

        # Downstream of C should be C
        downstream_c = get_all_downstream_nodes_recursive(self.graph1, "C")
        self.assertSetEqual(downstream_c, {"C"})

        # Downstream of D should be D
        downstream_d = get_all_downstream_nodes_recursive(self.graph1, "D")
        self.assertSetEqual(downstream_d, {"D"})

        # Non-existent node
        downstream_non_existent = get_all_downstream_nodes_recursive(self.graph1, "Z")
        self.assertSetEqual(downstream_non_existent, set())

        # Empty graph
        downstream_empty = get_all_downstream_nodes_recursive(self.empty_graph, "A")
        self.assertSetEqual(downstream_empty, set())

    def test_get_all_ancestors(self):
        # For self.graph1 (A -> B -> C, A -> D)
        ancestors_c = get_all_ancestors(self.graph1, "C")  # Ancestors of C are B, A
        self.assertSetEqual(ancestors_c, {"A", "B"})

        ancestors_b = get_all_ancestors(self.graph1, "B")  # Ancestors of B is A
        self.assertSetEqual(ancestors_b, {"A"})

        ancestors_d = get_all_ancestors(self.graph1, "D")  # Ancestors of D is A
        self.assertSetEqual(ancestors_d, {"A"})

        ancestors_a = get_all_ancestors(self.graph1, "A")  # A has no ancestors
        self.assertSetEqual(ancestors_a, set())

        # Non-existent node
        ancestors_non_existent = get_all_ancestors(self.graph1, "Z")
        self.assertSetEqual(ancestors_non_existent, set())

        # Empty graph
        ancestors_empty = get_all_ancestors(self.empty_graph, "A")
        self.assertSetEqual(ancestors_empty, set())

    def test_find_all_paths(self):
        # For self.graph1 (A -> B -> C, A -> D)
        paths_a_c = find_all_paths(self.graph1, "A", "C")
        self.assertListEqual(paths_a_c, [["A", "B", "C"]])

        paths_a_d = find_all_paths(self.graph1, "A", "D")
        self.assertListEqual(paths_a_d, [["A", "D"]])

        paths_a_b = find_all_paths(self.graph1, "A", "B")
        self.assertListEqual(paths_a_b, [["A", "B"]])

        paths_b_d = find_all_paths(self.graph1, "B", "D")  # No path
        self.assertListEqual(paths_b_d, [])

        paths_a_a = find_all_paths(self.graph1, "A", "A")  # Path to self
        self.assertListEqual(paths_a_a, [["A"]])

        # Diamond graph: S -> T1 -> E, S -> T2 -> E
        s, t1, t2, e = (Node("S", NodeType.COMPUTE), Node("T1", NodeType.COMPUTE, dependencies=["S"]), Node("T2", NodeType.COMPUTE, dependencies=["S"]), Node("E", NodeType.COMPUTE, dependencies=["T1", "T2"]))
        diamond_graph = TaskGraph("diamond")
        diamond_graph.add_nodes([s, t1, t2, e])
        paths_s_e = find_all_paths(diamond_graph, "S", "E")
        self.assertEqual(len(paths_s_e), 2)
        self.assertIn(["S", "T1", "E"], paths_s_e)
        self.assertIn(["S", "T2", "E"], paths_s_e)

        # Non-existent start/end
        self.assertListEqual(find_all_paths(self.graph1, "Z", "C"), [])
        self.assertListEqual(find_all_paths(self.graph1, "A", "Z"), [])


class TestTaskSplittingLogic(unittest.TestCase):
    """Tests for individual splitting strategy functions."""

    def assertGraphStructure(self, graph: TaskGraph, expected_node_ids: Set[str], expected_dependencies: Dict[str, List[str]], msg_prefix=""):
        self.assertSetEqual(set(graph.nodes.keys()), expected_node_ids, f"{msg_prefix} Node ID mismatch")
        graph.build_adjacency_lists()  # Ensure rev_adj is up-to-date
        for node_id, deps in expected_dependencies.items():
            self.assertIn(node_id, graph.rev_adj, f"{msg_prefix} Node {node_id} not in rev_adj")
            self.assertListEqual(sorted(graph.rev_adj[node_id]), sorted(deps), f"{msg_prefix} Dependencies for {node_id} mismatch")

    def test_split_single_structure_reconvergence(self):
        # Graph: A->B, F->C, then B,C -> D -> E (reconverge at D)
        a = Node("A", NodeType.DATA_LOAD)
        b = Node("B", NodeType.COMPUTE, dependencies=["A"])
        f = Node("F", NodeType.DATA_LOAD)
        c = Node("C", NodeType.COMPUTE, dependencies=["F"])
        d = Node("D", NodeType.COMPUTE, dependencies=["B", "C"])
        e_node = Node("E", NodeType.MODEL_TRAIN, dependencies=["D"])

        src_graph = TaskGraph("reconverge_src")
        src_graph.add_nodes([a, b, f, c, d, e_node])

        parallel_branches = [["A", "B"], ["F", "C"]]  # Node lists for branches up to merge point
        merge_node_id = "D"

        subgraphes = split_single_structure(src_graph, parallel_branches, merge_node_id, "base_idx")
        self.assertEqual(len(subgraphes), 2)

        # Expected subgraph 1: A->B->D->E
        # Expected subgraph 2: F->C->D->E
        subgraph1 = TaskGraph("exp1")
        subgraph1.add_nodes(
            [
                copy.deepcopy(a),
                copy.deepcopy(b),
                Node("D", NodeType.COMPUTE, dependencies=["B"]),  # D's deps adjusted
                Node("E", NodeType.MODEL_TRAIN, dependencies=["D"]),
            ]
        )
        subgraph2 = TaskGraph("exp2")
        subgraph2.add_nodes(
            [
                copy.deepcopy(f),
                copy.deepcopy(c),
                Node("D", NodeType.COMPUTE, dependencies=["C"]),  # D's deps adjusted
                Node("E", NodeType.MODEL_TRAIN, dependencies=["D"]),
            ]
        )
        expected_sig_1 = generate_structural_signature(subgraph1)
        expected_sig_2 = generate_structural_signature(subgraph2)

        actual_sigs = get_signatures(subgraphes)
        self.assertIn(expected_sig_1, actual_sigs)
        self.assertIn(expected_sig_2, actual_sigs)

    def test_split_by_fan_out_to_exits(self):
        # Graph: S -> A -> E1 (exit1)
        #        S -> B -> E2 (exit2)
        s = Node("S", NodeType.DATA_LOAD)
        a = Node("A", NodeType.COMPUTE, dependencies=["S"])
        e1 = Node("E1", NodeType.MODEL_TRAIN, dependencies=["A"])  # Exit 1
        b = Node("B", NodeType.COMPUTE, dependencies=["S"])
        e2 = Node("E2", NodeType.MODEL_TRAIN, dependencies=["B"])  # Exit 2

        src_graph = TaskGraph("fanout_src")
        src_graph.add_nodes([s, a, e1, b, e2])

        subgraphs = split_by_fan_out_to_exits(src_graph, 1)
        self.assertEqual(len(subgraphs), 2, "Should split into two subgraphs for distinct exits.")

        # Expected subgraph 1: S->A->E1
        # Expected subgraph 2: S->B->E2
        # Need to reconstruct expected TaskGraphs to get their signatures

        # Graph S->A->E1
        exp_g1 = TaskGraph("exp_g1")
        exp_g1.add_nodes([copy.deepcopy(s), Node("A", NodeType.COMPUTE, dependencies=["S"]), Node("E1", NodeType.MODEL_TRAIN, dependencies=["A"])])
        exp_g1_sig = generate_structural_signature(exp_g1)

        # Graph S->B->E2
        exp_g2 = TaskGraph("exp_g2")
        exp_g2.add_nodes([copy.deepcopy(s), Node("B", NodeType.COMPUTE, dependencies=["S"]), Node("E2", NodeType.MODEL_TRAIN, dependencies=["B"])])
        exp_g2_sig = generate_structural_signature(exp_g2)

        actual_sigs = get_signatures(subgraphs)
        self.assertIn(exp_g1_sig, actual_sigs)
        self.assertIn(exp_g2_sig, actual_sigs)

    def test_split_by_reconverging_paths_diamond(self):
        # Diamond: A -> B \
        #               -> D
        #          A -> C /
        a_ = Node("A", NodeType.DATA_LOAD)
        b_ = Node("B", NodeType.COMPUTE, dependencies=["A"])
        c_ = Node("C", NodeType.COMPUTE, dependencies=["A"])
        d_ = Node("D", NodeType.COMPUTE, dependencies=["B", "C"])
        src_graph = TaskGraph("diamond_src")
        src_graph.add_nodes([a_, b_, c_, d_])

        subgraphs = split_by_reconverging_paths(src_graph, 1)
        self.assertEqual(len(subgraphs), 2, "Diamond graph should split into two paths.")

        # Expected: A->B->D and A->C->D
        g_abd = TaskGraph("exp_abd")
        g_abd.add_nodes([copy.deepcopy(a_), Node("B", NodeType.COMPUTE, dependencies=["A"]), Node("D", NodeType.COMPUTE, dependencies=["B"])])
        g_acd = TaskGraph("exp_acd")
        g_acd.add_nodes([copy.deepcopy(a_), Node("C", NodeType.COMPUTE, dependencies=["A"]), Node("D", NodeType.COMPUTE, dependencies=["C"])])

        actual_sigs = get_signatures(subgraphs)
        self.assertIn(generate_structural_signature(g_abd), actual_sigs)
        self.assertIn(generate_structural_signature(g_acd), actual_sigs)


# Helper to extend TaskGraph for easier test setup
def add_nodes_and_build(self, nodes: List[Node]) -> TaskGraph:
    self.add_nodes(nodes)
    self.build_adjacency_lists()
    return self


TaskGraph.add_nodes_and_build = add_nodes_and_build


class TestDiscoverAndSplitParallelPaths(unittest.TestCase):
    """Comprehensive tests for the main discover_and_split_parallel_paths function."""

    def assertListOfGraphStructuresEqual(self, actual_graphs: List[TaskGraph], expected_graph_defs: List[Tuple[str, List[Node]]], msg=None):
        """
        Compares a list of actual TaskGraphs with a list of expected graph definitions.
        A graph definition is (name_for_debug, list_of_nodes_for_expected_graph).
        """
        self.assertEqual(len(actual_graphs), len(expected_graph_defs), f"{msg}: Different number of graphs. Got {len(actual_graphs)}, expected {len(expected_graph_defs)}")

        expected_signatures = []
        for name, nodes in expected_graph_defs:
            g = TaskGraph(name)
            g.add_nodes(nodes)
            expected_signatures.append(generate_structural_signature(g))

        actual_signatures = get_signatures(actual_graphs)
        self.assertListEqual(sorted(actual_signatures), sorted(expected_signatures), f"{msg}: Graph structures differ.")

    def test_empty_graph(self):
        empty_g = TaskGraph("empty")
        split_graphs = discover_and_split_parallel_paths(empty_g)
        self.assertEqual(len(split_graphs), 0, "Empty graph should result in empty list.")

    def test_linear_graph(self):
        # L1 -> L2 -> L3
        l1, l2, l3 = (Node("L1", NodeType.DATA_LOAD), Node("L2", NodeType.COMPUTE, dependencies=["L1"]), Node("L3", NodeType.MODEL_TRAIN, dependencies=["L2"]))
        linear_g = TaskGraph("linear")
        linear_g.add_nodes([l1, l2, l3])

        split_graphs = discover_and_split_parallel_paths(linear_g)

        self.assertListOfGraphStructuresEqual(
            split_graphs,
            [("linear_expected", [copy.deepcopy(n) for n in [l1, l2, l3]])],  # Deepcopy nodes for expected structure
        )

    def test_simple_reconvergence_diamond_graph(self):
        #   A
        #  / \
        # B   C
        #  \ /
        #   D
        a, b, c, d = (Node("A", NodeType.DATA_LOAD), Node("B", NodeType.COMPUTE, dependencies=["A"]), Node("C", NodeType.COMPUTE, dependencies=["A"]), Node("D", NodeType.COMPUTE, dependencies=["B", "C"]))
        diamond_g = TaskGraph("diamond")
        diamond_g.add_nodes([a, b, c, d])
        split_graphs = discover_and_split_parallel_paths(diamond_g)

        # Expected: (A->B->D) and (A->C->D)
        exp_nodes1 = [Node("A", NodeType.DATA_LOAD), Node("B", NodeType.COMPUTE, dependencies=["A"]), Node("D", NodeType.COMPUTE, dependencies=["B"])]
        exp_nodes2 = [Node("A", NodeType.DATA_LOAD), Node("C", NodeType.COMPUTE, dependencies=["A"]), Node("D", NodeType.COMPUTE, dependencies=["C"])]

        self.assertListOfGraphStructuresEqual(split_graphs, [("path_abd", exp_nodes1), ("path_acd", exp_nodes2)])

    def test_fan_out_only_graph(self):
        # A -> B (exit1)
        #   -> C (exit2)
        s_a, s_b_exit1, s_c_exit2 = Node("S_A", NodeType.DATA_LOAD), Node("S_B_exit1", NodeType.COMPUTE, dependencies=["S_A"]), Node("S_C_exit2", NodeType.COMPUTE, dependencies=["S_A"])
        fanout_g = TaskGraph("fanout_only")
        fanout_g.add_nodes([s_a, s_b_exit1, s_c_exit2])
        split_graphs = discover_and_split_parallel_paths(fanout_g)

        exp_nodes1 = [Node("S_A", NodeType.DATA_LOAD), Node("S_B_exit1", NodeType.COMPUTE, dependencies=["S_A"])]
        exp_nodes2 = [Node("S_A", NodeType.DATA_LOAD), Node("S_C_exit2", NodeType.COMPUTE, dependencies=["S_A"])]

        self.assertListOfGraphStructuresEqual(split_graphs, [("path_sa_sb", exp_nodes1), ("path_sa_sc", exp_nodes2)])

    def test_ex1_reconverge_from_prompt(self):
        # A -> B \
        #         -> C -> D_ex1 -> E_ex1
        # A1-> B1/
        node_a_orig = Node(node_id="A", node_type=NodeType.DATA_LOAD)
        node_b_orig = Node(node_id="B", node_type=NodeType.COMPUTE, dependencies=["A"])
        node_a1_orig = Node(node_id="A1", node_type=NodeType.DATA_LOAD)
        node_b1_orig = Node(node_id="B1", node_type=NodeType.COMPUTE, dependencies=["A1"])
        node_c_orig = Node(node_id="C", node_type=NodeType.COMPUTE, dependencies=["B", "B1"])
        node_d_ex1_orig = Node(node_id="D_ex1", node_type=NodeType.COMPUTE, dependencies=["C"])
        node_e_ex1_orig = Node(node_id="E_ex1", node_type=NodeType.MODEL_TRAIN, dependencies=["D_ex1"])

        original_graph_ex1 = TaskGraph(graph_id="ex1_reconverge")
        original_graph_ex1.add_nodes([node_a_orig, node_b_orig, node_a1_orig, node_b1_orig, node_c_orig, node_d_ex1_orig, node_e_ex1_orig])

        split_graphs = discover_and_split_parallel_paths(original_graph_ex1)

        # Expected path 1: A -> B -> C -> D_ex1 -> E_ex1
        exp1_nodes = [Node("A", NodeType.DATA_LOAD), Node("B", NodeType.COMPUTE, dependencies=["A"]), Node("C", NodeType.COMPUTE, dependencies=["B"]), Node("D_ex1", NodeType.COMPUTE, dependencies=["C"]), Node("E_ex1", NodeType.MODEL_TRAIN, dependencies=["D_ex1"])]
        # Expected path 2: A1 -> B1 -> C -> D_ex1 -> E_ex1
        exp2_nodes = [Node("A1", NodeType.DATA_LOAD), Node("B1", NodeType.COMPUTE, dependencies=["A1"]), Node("C", NodeType.COMPUTE, dependencies=["B1"]), Node("D_ex1", NodeType.COMPUTE, dependencies=["C"]), Node("E_ex1", NodeType.MODEL_TRAIN, dependencies=["D_ex1"])]
        self.assertListOfGraphStructuresEqual(split_graphs, [("path1_ex1", exp1_nodes), ("path2_ex1", exp2_nodes)], msg="TestEx1Reconverge")

    def test_ex2_complex_from_prompt(self):
        #      X -> P1 \
        #               -> M1 -> Z -> J1 -> K1 (exit1)
        #      Y -> P2 /        |
        #                       -> J2 -> K2 (exit2)
        #      P3 ---------------^ (P3 connects to Z)
        nx = Node("X", NodeType.DATA_LOAD)
        ny = Node("Y", NodeType.DATA_LOAD)
        np1 = Node("P1", NodeType.COMPUTE, dependencies=["X"])
        np2 = Node("P2", NodeType.COMPUTE, dependencies=["Y"])
        nm1 = Node("M1", NodeType.COMPUTE, dependencies=["P1", "P2"])
        np3 = Node("P3", NodeType.DATA_LOAD)
        nz = Node("Z", NodeType.COMPUTE, dependencies=["M1", "P3"])
        nj1 = Node("J1", NodeType.COMPUTE, dependencies=["Z"])
        nj2 = Node("J2", NodeType.COMPUTE, dependencies=["Z"])
        nk1 = Node("K1", NodeType.MODEL_TRAIN, dependencies=["J1"])  # Exit1
        nk2 = Node("K2", NodeType.MODEL_TRAIN, dependencies=["J2"])  # Exit2

        complex_graph = TaskGraph("ex2_complex")
        complex_graph.add_nodes([nx, ny, np1, np2, nm1, np3, nz, nj1, nj2, nk1, nk2])
        split_graphs = discover_and_split_parallel_paths(complex_graph)

        # Expected subgraphs (6 of them after full decomposition):
        # Path Group 1 (to K1):
        # 1. X->P1->M1->Z->J1->K1
        exp1_k1_nodes = [
            Node("X", NodeType.DATA_LOAD),
            Node("P1", NodeType.COMPUTE, dependencies=["X"]),
            Node("M1", NodeType.COMPUTE, dependencies=["P1"]),
            Node("Z", NodeType.COMPUTE, dependencies=["M1"]),
            Node("J1", NodeType.COMPUTE, dependencies=["Z"]),
            Node("K1", NodeType.MODEL_TRAIN, dependencies=["J1"]),
        ]
        # 2. Y->P2->M1->Z->J1->K1
        exp2_k1_nodes = [
            Node("Y", NodeType.DATA_LOAD),
            Node("P2", NodeType.COMPUTE, dependencies=["Y"]),
            Node("M1", NodeType.COMPUTE, dependencies=["P2"]),
            Node("Z", NodeType.COMPUTE, dependencies=["M1"]),
            Node("J1", NodeType.COMPUTE, dependencies=["Z"]),
            Node("K1", NodeType.MODEL_TRAIN, dependencies=["J1"]),
        ]
        # 3. P3->Z->J1->K1
        exp3_k1_nodes = [Node("P3", NodeType.DATA_LOAD), Node("Z", NodeType.COMPUTE, dependencies=["P3"]), Node("J1", NodeType.COMPUTE, dependencies=["Z"]), Node("K1", NodeType.MODEL_TRAIN, dependencies=["J1"])]

        # Path Group 2 (to K2):
        # 4. X->P1->M1->Z->J2->K2
        exp1_k2_nodes = [
            Node("X", NodeType.DATA_LOAD),
            Node("P1", NodeType.COMPUTE, dependencies=["X"]),
            Node("M1", NodeType.COMPUTE, dependencies=["P1"]),
            Node("Z", NodeType.COMPUTE, dependencies=["M1"]),
            Node("J2", NodeType.COMPUTE, dependencies=["Z"]),
            Node("K2", NodeType.MODEL_TRAIN, dependencies=["J2"]),
        ]
        # 5. Y->P2->M1->Z->J2->K2
        exp2_k2_nodes = [
            Node("Y", NodeType.DATA_LOAD),
            Node("P2", NodeType.COMPUTE, dependencies=["Y"]),
            Node("M1", NodeType.COMPUTE, dependencies=["P2"]),
            Node("Z", NodeType.COMPUTE, dependencies=["M1"]),
            Node("J2", NodeType.COMPUTE, dependencies=["Z"]),
            Node("K2", NodeType.MODEL_TRAIN, dependencies=["J2"]),
        ]
        # 6. P3->Z->J2->K2
        exp3_k2_nodes = [Node("P3", NodeType.DATA_LOAD), Node("Z", NodeType.COMPUTE, dependencies=["P3"]), Node("J2", NodeType.COMPUTE, dependencies=["Z"]), Node("K2", NodeType.MODEL_TRAIN, dependencies=["J2"])]

        self.assertListOfGraphStructuresEqual(split_graphs, [("xp1m1zj1k1", exp1_k1_nodes), ("yp2m1zj1k1", exp2_k1_nodes), ("p3zj1k1", exp3_k1_nodes), ("xp1m1zj2k2", exp1_k2_nodes), ("yp2m1zj2k2", exp2_k2_nodes), ("p3zj2k2", exp3_k2_nodes)], msg="TestEx2Complex")


if __name__ == "__main__":
    unittest.main(argv=["first-arg-is-ignored"], exit=False, verbosity=2)
